#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import logging
import os
from collections.abc import Iterable
from typing import Set, Union

import numpy as np
import torch
import torchvision.transforms as T
from PIL import Image
from transformers import RTDetrForObjectDetection, RTDetrImageProcessor

_log = logging.getLogger(__name__)


class LayoutPredictor:
    """
    Document layout prediction using safe tensors
    """

    def __init__(
        self,
        artifact_path: str,
        device: str = "cpu",
        num_threads: int = 4,
        base_threshold: float = 0.3,
        blacklist_classes: Set[str] = set(),
    ):
        """
        Provide the artifact path that contains the LayoutModel file

        Parameters
        ----------
        artifact_path: Path for the model torch file.
        device: (Optional) device to run the inference.
        num_threads: (Optional) Number of threads to run the inference if device = 'cpu'

        Raises
        ------
        FileNotFoundError when the model's torch file is missing
        """
        # Initialize classes map:
        self._classes_map = {
            0: "background",
            1: "Caption",
            2: "Footnote",
            3: "Formula",
            4: "List-item",
            5: "Page-footer",
            6: "Page-header",
            7: "Picture",
            8: "Section-header",
            9: "Table",
            10: "Text",
            11: "Title",
            12: "Document Index",
            13: "Code",
            14: "Checkbox-Selected",
            15: "Checkbox-Unselected",
            16: "Form",
            17: "Key-Value Region",
        }

        # Blacklisted classes
        self._black_classes = blacklist_classes  # set(["Form", "Key-Value Region"])

        # Set basic params
        self._threshold = base_threshold  # Score threshold
        self._image_size = 640
        self._size = np.asarray([[self._image_size, self._image_size]], dtype=np.int64)

        # Set number of threads for CPU
        self._device = torch.device(device)
        self._num_threads = num_threads
        if device == "cpu":
            torch.set_num_threads(self._num_threads)

        # Model file and configurations
        self._st_fn = os.path.join(artifact_path, "model.safetensors")
        if not os.path.isfile(self._st_fn):
            raise FileNotFoundError("Missing safe tensors file: {}".format(self._st_fn))

        # Load model and move to device
        processor_config = os.path.join(artifact_path, "preprocessor_config.json")
        model_config = os.path.join(artifact_path, "config.json")
        self._image_processor = RTDetrImageProcessor.from_json_file(processor_config)
        self._model = RTDetrForObjectDetection.from_pretrained(
            artifact_path, config=model_config
        ).to(self._device)
        self._model.eval()

        _log.debug("LayoutPredictor settings: {}".format(self.info()))

    def info(self) -> dict:
        """
        Get information about the configuration of LayoutPredictor
        """
        info = {
            "safe_tensors_file": self._st_fn,
            "device": self._device.type,
            "num_threads": self._num_threads,
            "image_size": self._image_size,
            "threshold": self._threshold,
        }
        return info

    @torch.inference_mode()
    def predict(self, orig_img: Union[Image.Image, np.ndarray]) -> Iterable[dict]:
        """
        Predict bounding boxes for a given image.
        The origin (0, 0) is the top-left corner and the predicted bbox coords are provided as:
        [left, top, right, bottom]

        Parameter
        ---------
        origin_img: Image to be predicted as a PIL Image object or numpy array.

        Yield
        -----
        Bounding box as a dict with the keys: "label", "confidence", "l", "t", "r", "b"

        Raises
        ------
        TypeError when the input image is not supported
        """
        # Convert image format
        if isinstance(orig_img, Image.Image):
            page_img = orig_img.convert("RGB")
        elif isinstance(orig_img, np.ndarray):
            page_img = Image.fromarray(orig_img).convert("RGB")
        else:
            raise TypeError("Not supported input image format")

        resize = {"height": self._image_size, "width": self._image_size}
        inputs = self._image_processor(
            images=page_img,
            return_tensors="pt",
            size=resize,
        ).to(self._device)
        outputs = self._model(**inputs)
        results = self._image_processor.post_process_object_detection(
            outputs,
            target_sizes=torch.tensor([page_img.size[::-1]]),
            threshold=self._threshold,
        )

        w, h = page_img.size

        result = results[0]
        for score, label_id, box in zip(
            result["scores"], result["labels"], result["boxes"]
        ):
            score = float(score.item())

            label_id = int(label_id.item()) + 1  # Advance the label_id
            label_str = self._classes_map[label_id]

            # Filter out blacklisted classes
            if label_str in self._black_classes:
                continue

            bbox_float = [float(b.item()) for b in box]
            l = min(w, max(0, bbox_float[0]))
            t = min(h, max(0, bbox_float[1]))
            r = min(w, max(0, bbox_float[2]))
            b = min(h, max(0, bbox_float[3]))
            yield {
                "l": l,
                "t": t,
                "r": r,
                "b": b,
                "label": label_str,
                "confidence": score,
            }



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import os
import platform
import re
from typing import Dict, Union


class MemMonitor:
    r"""
    Memory monitor for Linux

    It supports 2 approaches for extracting memory information:
    - linux-native: It parse the `/proc` pseudo-files. It is available only for Linux
    - psutil: Use the `psutil` library

    ## Linux-Native approach

    The linux-native approach implements 2 methods to extract the memory fields:

    1. The `get_memory()` method:

    - It is very fast
    - It parses the `/proc/<pid>/statm` pseudo-file
    - It Contains the following fields:
        size       (1) total program size
                   (same as VmSize in /proc/[pid]/status)
        resident   (2) resident set size
                   (same as VmRSS in /proc/[pid]/status)
        shared     (3) number of resident shared pages (i.e., backed by a file)
                   (same as RssFile+RssShmem in /proc/[pid]/status)
        text       (4) text (code)
        lib        (5) library (unused since Linux 2.6; always 0)
        data       (6) data + stack
        dt         (7) dirty pages (unused since Linux 2.6; always 0)


    2. The `get_memory_full()` method:

    - It is slower to parse but contains more detailed information
    - It uses regex to parse the `/proc/<pid>/status` pseudo-file
    - It contains the following fields:
        VmPeak: Peak virtual memory size.
        VmSize: Virtual memory size.
        VmLck: Locked memory size (see mlock(2)).
        VmPin: Pinned memory size (since Linux 3.2). These are pages that can't be moved because
               something needs to directly access physical memory.
        VmHWM: Peak resident set size ("high water mark").
        VmRSS: Resident set size.  Note that the value here is the sum of RssAnon, RssFile, and
               RssShmem.
        RssAnon: Size of resident anonymous memory.  (since Linux 4.5).
        RssFile: Size of resident file mappings.  (since Linux 4.5).
        RssShmem: Size of resident shared memory (includes System V shared memory, mappings from
                  tmpfs(5), and shared anonymous mappings).  (since Linux 4.5).
        VmData, VmStk, VmExe: Size of data, stack, and text segments.
        VmLib: Shared library code size.
        VmPTE: Page table entries size (since Linux 2.6.10).
        VmPMD: Size of second-level page tables (added in Linux 4.0; removed in Linux 4.15).
        VmSwap: Swapped-out virtual memory size by anonymous private pages; shmem swap usage is
                not included (since Linux 2.6.34).


    ## The psutil library

    - Apparently the psutil library parses the `/proc/<pid>/statm`
    - The memory_info() function returns the fields: rss, vms, shared, text, lib, data, dirty


    ## Field mappings

    These are the fields returned by psutil memory_info() and their mapping in the /proc files:
    (I put ? when I am not 100% about the mapping)

    | psutil  | /proc/$$/status    | /proc/$$/statm |
    |---------|--------------------|----------------|
    | rss     | VmRSS              | resident       |
    | vms     | VmSize             | size           |
    | shared  | RssFile + RssShmem | shared         |
    | text    | VmExe ?            | text           |
    | lib     | RssShmem ?         | lib            |
    | data    | VmData + VmStk     | data           |
    | dirty   | VmSwap ?           | dt             |

    """

    def __init__(self, enable=True):
        self._enable = enable
        self._pid = os.getpid()

        # Create regex for each memory field of the /proc/status pseudo-file
        self._status_fields = [
            "VmPeak",
            "VmSize",
            "VmLck",
            "VmPin",
            "VmHWM",
            "VmRSS",
            "RssAnon",
            "RssFile",
            "RssShmem",
            "VmData",
            "VmStk",
            "VmExe",
            "VmLib",
            "VmPTE",
            "VmPMD",
            "VmSwap",
        ]
        self._status_regex = {}
        for mem_field in self._status_fields:
            regex_str = r"({}:)(\s+)(\d*)(.*)".format(mem_field)
            self._status_regex[mem_field] = re.compile(regex_str)

    def get_memory_full(self) -> Union[Dict, int]:
        r"""
        - Parse /proc/<pid>status to get all memory info.
        - The method returns a dict with the fields self._status_fields
        - This method is SLOW. Unless you need the full memory info, better to use `get_memory`

        The returned values are in kB
        """
        if not self._enable:
            return -2
        if platform.system() != "Linux":
            return -1
        pid_fn = "/proc/{}/status".format(self._pid)

        # Dict to collect all memory fields
        memory = {}
        with open(pid_fn, "r") as fn:
            for ll in fn:
                for mem_field in self._status_fields:
                    regex = self._status_regex[mem_field]
                    m = regex.match(ll)
                    if m is not None:
                        memory[mem_field] = int(m.group(3))
                if len(memory) == len(self._status_fields):
                    break

        return memory

    def get_memory(self) -> Union[Dict, int]:
        r"""
        - Parse /proc/<pid>statm to get the most important memory fields
        - This is a fast implementation.
        - The method returns a dict with the fields:
            "size", "resident", "shared", "text", "lib", "data", "dt"
        - Check the documentation at the top for a mapping across the various fields

        The returned values are in kB
        """
        if not self._enable:
            return -2
        if platform.system() != "Linux":
            return -1
        pid_fn = "/proc/{}/statm".format(self._pid)

        # Dict to collect all memory fields
        memory = {}
        with open(pid_fn, "r") as fn:
            ll = fn.read()
            # The values are in pages
            # Each page is 4096 bytes (4kB)
            data = [int(x) << 2 for x in ll.split(" ")]
            memory = {
                "size": data[0],
                "resident": data[1],
                "shared": data[2],
                "text": data[3],
                "lib": data[4],
                "data": data[5],
                "dt": data[6],
            }
        return memory



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import copy
import logging
from itertools import groupby

import docling_ibm_models.tableformer.settings as s

LOG_LEVEL = logging.INFO
# LOG_LEVEL = logging.DEBUG
logger = s.get_custom_logger("consolidate", LOG_LEVEL)
# png_files = {}  # Evaluation files
total_pics = 0


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def otsl_clean(rs_list):
    new_rs_list = []
    stop_list = ["<pad>", "<unk>", "<start>", "<end>"]
    for tag in rs_list:
        if tag not in stop_list:
            new_rs_list.append(tag)
    return new_rs_list


def otsl_sqr_chk(rs_list, name, logdebug):
    rs_list_split = [
        list(group) for k, group in groupby(rs_list, lambda x: x == "nl") if not k
    ]
    isSquare = True
    if len(rs_list_split) > 0:
        init_tag_len = len(rs_list_split[0]) + 1
        for ind, ln in enumerate(rs_list_split):
            ln.append("nl")
            if len(ln) != init_tag_len:
                isSquare = False
        if isSquare:
            if logdebug:
                logger.debug(
                    "{}*OK* Table is square! *OK*{}".format(
                        bcolors.OKGREEN, bcolors.ENDC
                    )
                )
        else:
            err_name = "{}*ERR* " + name + " *ERR*{}"
            logger.debug(err_name.format(bcolors.FAIL, bcolors.ENDC))
            logger.debug(
                "{}*ERR* Table is not square! *ERR*{}".format(
                    bcolors.FAIL, bcolors.ENDC
                )
            )
    return isSquare


def otsl_pad_to_sqr(rs_list, pad_tag):
    new_list = []
    rs_list_split = [
        list(group) for k, group in groupby(rs_list, lambda x: x == "nl") if not k
    ]
    max_row_len = 0
    for ind, ln in enumerate(rs_list_split):
        if len(ln) > max_row_len:
            max_row_len = len(ln)
    for ind, ln in enumerate(rs_list_split):
        ln += [pad_tag] * (max_row_len - len(ln))
        ln.append("nl")
        new_list.extend(ln)
    return new_list


def otsl_tags_cells_sync_chk(rs_list, cells, name, logdebug):
    countCellTags = 0
    isGood = True
    for rsTag in rs_list:
        if rsTag in ["fcel", "ched", "rhed", "srow", "ecel"]:
            countCellTags += 1
    if countCellTags != len(cells):
        err_name = "{}*!ERR* " + name + " *ERR!*{}"
        logger.debug(err_name.format(bcolors.FAIL, bcolors.ENDC))
        err_msg = "{}*!ERR* Tags are not in sync with cells! *ERR!*{}"
        logger.debug(err_msg.format(bcolors.FAIL, bcolors.ENDC))
        isGood = False
    return isGood


def otsl_check_down(rs_split, x, y):
    distance = 1
    elem = "ucel"
    goodlist = ["fcel", "ched", "rhed", "srow", "ecel", "lcel", "nl"]
    while elem not in goodlist and y < len(rs_split) - 1:
        y += 1
        distance += 1
        elem = rs_split[y][x]
    if elem in goodlist:
        distance -= 1
    return distance


def otsl_check_right(rs_split, x, y):
    distance = 1
    elem = "lcel"
    goodlist = ["fcel", "ched", "rhed", "srow", "ecel", "ucel", "nl"]
    while elem not in goodlist and x < (len(rs_split[y]) - 1):
        x += 1
        distance += 1
        elem = rs_split[y][x]
    if elem in goodlist:
        distance -= 1
    return distance


def otsl_to_html(rs_list, logdebug):
    if len(rs_list) == 0:
        return []

    if rs_list[0] not in ["fcel", "ched", "rhed", "srow", "ecel"]:
        # Most likely already HTML...
        return rs_list
    html_table = []
    if logdebug:
        logger.debug(
            "{}*Reconstructing HTML...*{}".format(bcolors.WARNING, bcolors.ENDC)
        )

    if not otsl_sqr_chk(rs_list, "---", logdebug):
        # PAD TABLE TO SQUARE
        logger.debug("{}*Padding to square...*{}".format(bcolors.WARNING, bcolors.ENDC))
        rs_list = otsl_pad_to_sqr(rs_list, "lcel")

    # 2D structure, line by line:
    rs_list_split = [
        list(group) for k, group in groupby(rs_list, lambda x: x == "nl") if not k
    ]

    if logdebug:
        logger.debug("")

    # Sequentially store indexes of 2D spans that were registered to avoid re-registering them
    registry_2d_span = []

    # Iterate all elements in the rs line, and look right / down to detect spans
    # If span detected - run function to find size of the span
    # repeat with all cells
    thead_present = False

    for rs_row_ind, rs_row in enumerate(rs_list_split):
        html_list = []

        if not thead_present:
            if "ched" in rs_list_split[rs_row_ind]:
                html_list.append("<thead>")
                thead_present = True

        if thead_present:
            if "ched" not in rs_list_split[rs_row_ind]:
                html_list.append("</thead>")
                thead_present = False

        html_list.append("<tr>")
        for rs_cell_ind, rs_cell in enumerate(rs_list_split[rs_row_ind]):
            if rs_cell in ["fcel", "ched", "rhed", "srow", "ecel"]:
                rdist = 0
                ddist = 0
                xrdist = 0
                xddist = 0
                span = False
                # Check if it has horizontal span:
                if rs_cell_ind + 1 < len(rs_list_split[rs_row_ind]):
                    if rs_list_split[rs_row_ind][rs_cell_ind + 1] == "lcel":
                        rdist = otsl_check_right(rs_list_split, rs_cell_ind, rs_row_ind)
                        span = True
                # Check if it has vertical span:
                if rs_row_ind + 1 < len(rs_list_split):
                    # logger.debug(">>>")
                    # logger.debug(rs_list_split[rs_row_ind + 1])
                    # logger.debug(">>> rs_cell_ind = {}".format(rs_cell_ind))
                    if rs_list_split[rs_row_ind + 1][rs_cell_ind] == "ucel":
                        ddist = otsl_check_down(rs_list_split, rs_cell_ind, rs_row_ind)
                        span = True
                # Check if it has 2D span:
                if rs_cell_ind + 1 < len(rs_list_split[rs_row_ind]):
                    if rs_list_split[rs_row_ind][rs_cell_ind + 1] == "xcel":
                        xrdist = otsl_check_right(
                            rs_list_split, rs_cell_ind, rs_row_ind
                        )
                        xddist = otsl_check_down(rs_list_split, rs_cell_ind, rs_row_ind)
                        span = True
                        # Check if this 2D span was already registered,
                        # If not - register, if yes - cancel span
                        # logger.debug("rs_cell_ind: {}, xrdist:{}".format(rs_cell_ind, xrdist))
                        # logger.debug("rs_row_ind: {}, xddist:{}".format(rs_cell_ind, xrdist))
                        for x in range(rs_cell_ind, xrdist + rs_cell_ind):
                            for y in range(rs_row_ind, xddist + rs_row_ind):
                                reg2dind = str(x) + "_" + str(y)
                                # logger.debug(reg2dind)
                                if reg2dind in registry_2d_span:
                                    # Cell of the span is already in, cancel current span
                                    span = False
                        if span:
                            # None of the span cells were previously registered
                            # Register an entire span
                            for x in range(rs_cell_ind, xrdist + rs_cell_ind):
                                for y in range(rs_row_ind, xddist + rs_row_ind):
                                    reg2dind = str(x) + "_" + str(y)
                                    registry_2d_span.append(reg2dind)
                if span:
                    html_list.append("<td")
                    if rdist > 1:
                        html_list.append(' colspan="' + str(rdist) + '"')
                    if ddist > 1:
                        html_list.append(' rowspan="' + str(ddist) + '"')
                    if xrdist > 1:
                        html_list.append(' rowspan="' + str(xddist) + '"')
                        html_list.append(' colspan="' + str(xrdist) + '"')
                    html_list.append(">")
                    html_list.append("</td>")
                else:
                    html_list.append("<td>")
                    html_list.append("</td>")
        html_list.append("</tr>")
        html_table.extend(html_list)

    if logdebug:
        logger.debug(
            "*********************** registry_2d_span ***************************"
        )
        logger.debug(registry_2d_span)
        logger.debug(
            "********************************************************************"
        )

    return html_table


def html_to_otsl(table, writer, logdebug, extra_debug, include_html, use_writer):
    r"""
    Converts table structure from HTML to RS

    Parameters
    ----------
    table : json
        line from jsonl
    writer : writer
        Writes lines into output jsonl
    """

    table_html_structure = copy.deepcopy(table["html"]["structure"])
    out_line = table
    if include_html:
        out_line["html"]["html_structure"] = table_html_structure
        out_line["html"]["html_restored_structure"] = {"tokens": []}

    out_line["html"]["structure"] = {"tokens": []}
    # possible colspans
    pos_colspans = {
        ' colspan="20"': 20,
        ' colspan="19"': 19,
        ' colspan="18"': 18,
        ' colspan="17"': 17,
        ' colspan="16"': 16,
        ' colspan="15"': 15,
        ' colspan="14"': 14,
        ' colspan="13"': 13,
        ' colspan="12"': 12,
        ' colspan="11"': 11,
        ' colspan="10"': 10,
        ' colspan="2"': 2,
        ' colspan="3"': 3,
        ' colspan="4"': 4,
        ' colspan="5"': 5,
        ' colspan="6"': 6,
        ' colspan="7"': 7,
        ' colspan="8"': 8,
        ' colspan="9"': 9,
    }
    # possible rowspans
    pos_rowspans = {
        ' rowspan="20"': 20,
        ' rowspan="19"': 19,
        ' rowspan="18"': 18,
        ' rowspan="17"': 17,
        ' rowspan="16"': 16,
        ' rowspan="15"': 15,
        ' rowspan="14"': 14,
        ' rowspan="13"': 13,
        ' rowspan="12"': 12,
        ' rowspan="11"': 11,
        ' rowspan="10"': 10,
        ' rowspan="2"': 2,
        ' rowspan="3"': 3,
        ' rowspan="4"': 4,
        ' rowspan="5"': 5,
        ' rowspan="6"': 6,
        ' rowspan="7"': 7,
        ' rowspan="8"': 8,
        ' rowspan="9"': 9,
    }

    t_cells = []  # 2D structure
    tl_cells = []  # 1D structure
    t_expands = []  # 2D structure
    tl_spans = {}  # MAP, POPULATE WITH ACTUAL SPANS VALUES, IN SYNC WITH tl_cells

    current_line = 0
    current_column = 0
    current_html_cell_ind = 0

    current_line_tags = []
    current_line_expands = []

    if logdebug:
        logger.debug("")
        logger.debug("*** {}: {} ***".format(table["split"], table["filename"]))

    colnum = 0

    if extra_debug:
        logger.debug(
            "========================== Input HTML ============================"
        )
        logger.debug(table_html_structure["tokens"])
        logger.debug(
            "=================================================================="
        )

    if logdebug:
        logger.debug("********")
        logger.debug("* OTSL *")
        logger.debug("********")

    for i in range(len(table_html_structure["tokens"])):
        html_tag = table_html_structure["tokens"][i]
        prev_html_tag = ""
        next_html_tag = ""
        if i > 0:
            prev_html_tag = table_html_structure["tokens"][i - 1]
        if i < len(table_html_structure["tokens"]) - 1:
            next_html_tag = table_html_structure["tokens"][i + 1]

        if html_tag not in ["<thead>", "<tbody>"]:
            # Then check the next tag...
            # rules of conversion
            # Check up-cell in t_expands, in case row-spans have to be inserted
            if html_tag in ["<td>", "<td", "</tr>"]:
                if current_line > 0:
                    if current_column >= len(t_expands[current_line - 1]):
                        # !!!
                        return False, {}
                    up_expand = t_expands[current_line - 1][current_column]

                    while up_expand[1] > 0:
                        if up_expand[0] == 0:
                            # ucel
                            current_line_tags.append("ucel")
                            current_line_expands.append([0, up_expand[1] - 1])
                            current_column += 1
                        else:
                            # xcel
                            for ci in range(up_expand[0]):
                                current_line_tags.append("xcel")
                                current_line_expands.append(
                                    [up_expand[0] - ci, up_expand[1] - 1]
                                )
                                current_column += 1
                        up_expand = t_expands[current_line - 1][current_column]
            # ======================================================================================
            # Fix for trailing "ucel" in a row
            if html_tag in ["</tr>"]:
                if current_line > 0:
                    cur_line_len = len(current_line_expands)
                    pre_line_len = len(t_expands[current_line - 1])

                    if cur_line_len < pre_line_len:
                        extra_columns = pre_line_len - cur_line_len - 1
                        if extra_columns > 0:
                            if extra_debug:
                                logger.debug(
                                    "Extra columns needed in row: {}".format(
                                        extra_columns
                                    )
                                )

                            for clm in range(extra_columns):
                                up_expand = t_expands[current_line - 1][
                                    cur_line_len + clm
                                ]
                                if up_expand[0] == 0:
                                    # ucel
                                    current_line_tags.append("ucel")
                                    current_line_expands.append([0, up_expand[1] - 1])
                                else:
                                    # xcel
                                    current_line_tags.append("xcel")
                                    current_line_expands.append(
                                        [up_expand[0], up_expand[1] - 1]
                                    )
            # ======================================================================================

            # 1. Opening cell tags
            if html_tag in ["<td>", "<td"]:
                # check if cell is empty...
                cell_is_empty = True
                if "cells" in table["html"]:
                    cell_tokens = table["html"]["cells"][current_html_cell_ind][
                        "tokens"
                    ]
                else:
                    cell_tokens = "f"

                # Clean cell_tokens from trash:
                cell_tokens = list(filter(lambda a: a != "<i>", cell_tokens))
                cell_tokens = list(filter(lambda a: a != "<I>", cell_tokens))
                cell_tokens = list(filter(lambda a: a != "<b>", cell_tokens))
                cell_tokens = list(filter(lambda a: a != "<B>", cell_tokens))
                cell_tokens = list(filter(lambda a: a != " ", cell_tokens))
                cell_tokens = list(filter(lambda a: a != "</b>", cell_tokens))
                cell_tokens = list(filter(lambda a: a != "</B>", cell_tokens))
                cell_tokens = list(filter(lambda a: a != "</i>", cell_tokens))
                cell_tokens = list(filter(lambda a: a != "</I>", cell_tokens))

                # Check if empty
                if len(cell_tokens) > 0:
                    cell_is_empty = False
                if cell_is_empty:
                    out_line["html"]["cells"][current_html_cell_ind]["tokens"] = []
                    current_line_tags.append("ecel")
                    current_line_expands.append([0, 0])
                else:
                    current_line_tags.append("fcel")
                    current_line_expands.append([0, 0])
                current_html_cell_ind += 1
                current_column += 1

            # 2. Closing row tags
            if html_tag == "</tr>":
                if len(current_line_tags) > colnum:
                    colnum = len(current_line_tags)
                # Save everything we read about the line to t_cells
                current_line_tags.append("nl")
                t_cells.append(copy.deepcopy(current_line_tags))
                tl_cells.extend(copy.deepcopy(current_line_tags))
                if logdebug:
                    print(current_line_tags)
                current_line_tags = []

                # Deal with expands
                current_line_expands.append([-1, -1])
                # Output spans metadata
                t_expands.append(copy.deepcopy(current_line_expands))
                current_line_expands = []

                current_column = 0
                current_line += 1
            # 3. Colspans only
            if html_tag in pos_colspans:
                if prev_html_tag not in pos_rowspans:
                    if next_html_tag not in pos_rowspans:
                        colspan_len = pos_colspans[html_tag]
                        tl_spans[current_html_cell_ind - 1] = [colspan_len, 1]
                        current_line_expands[len(current_line_expands) - 1] = [
                            colspan_len,
                            0,
                        ]
                        for ci in range(colspan_len - 1):
                            current_line_tags.append("lcel")
                            current_line_expands.append([colspan_len - ci - 1, 0])
                            current_column += 1

            # 4. Rowspans only
            if html_tag in pos_rowspans:
                if prev_html_tag not in pos_colspans:
                    if next_html_tag not in pos_colspans:
                        rowspan_len = pos_rowspans[html_tag]
                        tl_spans[current_html_cell_ind - 1] = [1, rowspan_len]
                        current_line_expands[len(current_line_expands) - 1] = [
                            0,
                            rowspan_len - 1,
                        ]

            # 5. 2D spans
            if html_tag in pos_rowspans:
                rowspan_len = pos_rowspans[html_tag]
                if prev_html_tag in pos_colspans:
                    colspan_len = pos_colspans[prev_html_tag]
                    tl_spans[current_html_cell_ind - 1] = [colspan_len, rowspan_len]
                    newexp = [colspan_len, rowspan_len - 1]
                    current_line_expands[len(current_line_expands) - 1] = newexp
                    for ci in range(colspan_len - 1):
                        current_line_tags.append("xcel")
                        current_line_expands.append(
                            [colspan_len - ci - 1, rowspan_len - 1]
                        )
                if next_html_tag in pos_colspans:
                    colspan_len = pos_colspans[next_html_tag]
                    tl_spans[current_html_cell_ind - 1] = [colspan_len, rowspan_len]
                    newexp = [colspan_len, rowspan_len - 1]
                    current_line_expands[len(current_line_expands) - 1] = newexp
                    for ci in range(colspan_len - 1):
                        current_line_tags.append("xcel")
                        current_line_expands.append(
                            [colspan_len - ci - 1, rowspan_len - 1]
                        )

    t_name = "*** {}: {} ***".format(table["split"], table["filename"])
    # check if square
    isSquare = otsl_sqr_chk(tl_cells, t_name, logdebug)
    # TODO: pad if not square?
    if not isSquare:
        tl_cells = otsl_pad_to_sqr(tl_cells, "fcel")
    # check if cells (bboxes) in sync:
    if "cells" in out_line["html"]:
        isGood = otsl_tags_cells_sync_chk(
            tl_cells, out_line["html"]["cells"], t_name, logdebug
        )
    # convert back to HTML
    rHTML = []
    if isSquare:
        rHTML = otsl_to_html(tl_cells, logdebug)
    out_line["html"]["html_restored_structure"]["tokens"] = rHTML

    out_line["html"]["structure"]["tokens"] = tl_cells
    out_line["otsl_spans"] = tl_spans
    out_line["cols"] = colnum
    out_line["rows"] = len(t_cells)
    out_line["html_len"] = len(table_html_structure["tokens"])
    out_line["rs_len"] = len(tl_cells)
    # save converted line
    if use_writer:
        if isSquare:
            if isGood:
                writer.write(out_line)

    if logdebug:
        logger.debug("{}Reconstructed HTML:{}".format(bcolors.OKGREEN, bcolors.ENDC))
        logger.debug(rHTML)
        # original HTML
        oHTML = out_line["html"]["html_structure"]
        logger.debug("{}Original HTML:{}".format(bcolors.OKBLUE, bcolors.ENDC))
        logger.debug(oHTML)

    return True, out_line



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import glob
import json
import logging
import os
from itertools import groupby
from pathlib import Path

import cv2
import numpy as np
import torch
from safetensors.torch import load_model

import docling_ibm_models.tableformer.common as c
import docling_ibm_models.tableformer.data_management.transforms as T
import docling_ibm_models.tableformer.settings as s
import docling_ibm_models.tableformer.utils.utils as u
from docling_ibm_models.tableformer.data_management.matching_post_processor import (
    MatchingPostProcessor,
)
from docling_ibm_models.tableformer.data_management.tf_cell_matcher import CellMatcher
from docling_ibm_models.tableformer.models.common.base_model import BaseModel
from docling_ibm_models.tableformer.models.table04_rs.tablemodel04_rs import (
    TableModel04_rs,
)
from docling_ibm_models.tableformer.otsl import otsl_to_html
from docling_ibm_models.tableformer.utils.app_profiler import AggProfiler

# LOG_LEVEL = logging.INFO
# LOG_LEVEL = logging.DEBUG
LOG_LEVEL = logging.WARN

logger = s.get_custom_logger(__name__, LOG_LEVEL)


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def otsl_sqr_chk(rs_list, logdebug):
    rs_list_split = [
        list(group) for k, group in groupby(rs_list, lambda x: x == "nl") if not k
    ]
    isSquare = True
    if len(rs_list_split) > 0:
        init_tag_len = len(rs_list_split[0]) + 1

        totcelnum = rs_list.count("fcel") + rs_list.count("ecel")
        if logdebug:
            logger.debug("Total number of cells = {}".format(totcelnum))

        for ind, ln in enumerate(rs_list_split):
            ln.append("nl")
            if logdebug:
                logger.debug("{}".format(ln))
            if len(ln) != init_tag_len:
                isSquare = False
        if isSquare:
            if logdebug:
                logger.debug(
                    "{}*OK* Table is square! *OK*{}".format(
                        bcolors.OKGREEN, bcolors.ENDC
                    )
                )
        else:
            if logdebug:
                err_name = "{}***** ERR ******{}"
                logger.debug(err_name.format(bcolors.FAIL, bcolors.ENDC))
                logger.debug(
                    "{}*ERR* Table is not square! *ERR*{}".format(
                        bcolors.FAIL, bcolors.ENDC
                    )
                )
    return isSquare


class TFPredictor:
    r"""
    Table predictions for the in-memory Docling API
    """

    def __init__(self, config, device: str = "cpu", num_threads: int = 4):
        r"""
        Parameters
        ----------
        config : dict Parameters configuration
        device: (Optional) torch device to run the inference.
        num_threads: (Optional) Number of threads to run the inference if device = 'cpu'

        Raises
        ------
        ValueError
        When the model cannot be found
        """
        # self._device = torch.device(device)
        self._device = device
        self._log().info("Running on device: {}".format(device))

        self._config = config
        self.enable_post_process = True

        self._padding = config["predict"].get("padding", False)
        self._padding_size = config["predict"].get("padding_size", 10)

        self._cell_matcher = CellMatcher(config)
        self._post_processor = MatchingPostProcessor(config)

        self._init_word_map()

        # Set the number of threads
        if device == "cpu":
            self._num_threads = num_threads
            torch.set_num_threads(self._num_threads)

        # Load the model
        self._model = self._load_model()
        self._model.eval()
        self._prof = config["predict"].get("profiling", False)
        self._profiling_agg_window = config["predict"].get("profiling_agg_window", None)
        if self._profiling_agg_window is not None:
            AggProfiler(self._profiling_agg_window)
        else:
            AggProfiler()

    def _init_word_map(self):
        self._prepared_data_dir = c.safe_get_parameter(
            self._config, ["dataset", "prepared_data_dir"], required=False
        )

        if self._prepared_data_dir is None:
            self._word_map = c.safe_get_parameter(
                self._config, ["dataset_wordmap"], required=True
            )
        else:
            data_name = c.safe_get_parameter(
                self._config, ["dataset", "name"], required=True
            )
            word_map_fn = c.get_prepared_data_filename("WORDMAP", data_name)

            # Load word_map
            with open(os.path.join(self._prepared_data_dir, word_map_fn), "r") as f:
                self._log().debug("Load WORDMAP from: {}".format(word_map_fn))
                self._word_map = json.load(f)

        self._init_data = {"word_map": self._word_map}
        # Prepare a reversed index for the word map
        self._rev_word_map = {v: k for k, v in self._word_map["word_map_tag"].items()}

    def get_init_data(self):
        r"""
        Return the initialization data
        """
        return self._init_data

    def get_model(self):
        r"""
        Return the loaded model
        """
        return self._model

    def _load_model(self):
        r"""
        Load the proper model
        """

        self._model_type = self._config["model"]["type"]
        model = TableModel04_rs(self._config, self._init_data, self._device)

        if model is None:
            err_msg = "Not able to initiate a model for {}".format(self._model_type)
            self._log().error(err_msg)
            raise ValueError(err_msg)

        self._remove_padding = False
        if self._model_type == "TableModel02":
            self._remove_padding = True

        # Load model from safetensors
        save_dir = self._config["model"]["save_dir"]
        models_fn = glob.glob(f"{save_dir}/tableformer_*.safetensors")
        if not models_fn:
            err_msg = "Not able to find a model file for {}".format(self._model_type)
            self._log().error(err_msg)
            raise ValueError(err_msg)
        model_fn = models_fn[
            0
        ]  # Take the first tableformer safetensors file inside the save_dir
        missing, unexpected = load_model(model, model_fn, device=self._device)
        if missing or unexpected:
            err_msg = "Not able to load the model weights for {}".format(
                self._model_type
            )
            self._log().error(err_msg)
            raise ValueError(err_msg)

        return model

    def get_device(self):
        return self._device

    def get_model_type(self):
        return self._model_type

    def _log(self):
        # Setup a custom logger
        return s.get_custom_logger(self.__class__.__name__, LOG_LEVEL)

    def _deletebbox(self, listofbboxes, index):
        newlist = []
        for i in range(len(listofbboxes)):
            bbox = listofbboxes[i]
            if i not in index:
                newlist.append(bbox)
        return newlist

    def _remove_bbox_span_desync(self, prediction):
        # Delete 1 extra bbox after span tag
        index_to_delete_from = 0
        indexes_to_delete = []
        newbboxes = []
        for html_elem in prediction["html_seq"]:
            if html_elem == "<td>":
                index_to_delete_from += 1
            if html_elem == ">":
                index_to_delete_from += 1
                # remove element from bboxes
                self._log().debug(
                    "========= DELETE BBOX INDEX: {}".format(index_to_delete_from)
                )
                indexes_to_delete.append(index_to_delete_from)

        newbboxes = self._deletebbox(prediction["bboxes"], indexes_to_delete)
        return newbboxes

    def _check_bbox_sync(self, prediction):
        bboxes = []
        match = False
        # count bboxes
        count_bbox = len(prediction["bboxes"])
        # count td tags
        count_td = 0
        for html_elem in prediction["html_seq"]:
            if html_elem == "<td>" or html_elem == ">":
                count_td += 1
            if html_elem in ["fcel", "ecel", "ched", "rhed", "srow"]:
                count_td += 1
        self._log().debug(
            "======================= PREDICTED BBOXES: {}".format(count_bbox)
        )
        self._log().debug(
            "=======================  PREDICTED CELLS: {}".format(count_td)
        )
        if count_bbox != count_td:
            bboxes = self._remove_bbox_span_desync(prediction)
        else:
            bboxes = prediction["bboxes"]
            match = True
        return match, bboxes

    def page_coords_to_table_coords(self, bbox, table_bbox, im_width, im_height):
        r"""
        Transforms given bbox from page coordinate system into table image coordinate system

        Parameters
        ----------
        bbox : list
            bbox to transform in page coordinates
        table_bbox : list
            table bbox, in page coordinates
        im_width : integer
            width of an image with rendered table (in pixels)
        im_height : integer
            height of an image height rendered table (in pixels)

        Returns
        -------
        bbox: list
            bbox with transformed coordinates
        """
        # Coordinates of given bbox
        x1 = bbox[0]
        y1 = bbox[1]
        x2 = bbox[2]
        y2 = bbox[3]

        # Coordinates of table bbox
        t_x1 = table_bbox[0]
        t_y1 = table_bbox[1]
        t_x2 = table_bbox[2]
        t_y2 = table_bbox[3]

        # Table width / height
        tw = t_x2 - t_x1
        th = t_y2 - t_y1
        new_bbox = [0, 0, 0, 0]
        # Flip corners, substract table coordinates and rescale to new image size
        new_bbox[0] = im_width * (x1 - t_x1) / tw
        new_bbox[1] = im_height * (t_y2 - y2) / th
        new_bbox[2] = im_width * (x2 - t_x1) / tw
        new_bbox[3] = im_height * (t_y2 - y1) / th

        return new_bbox

    def _depad_bboxes(self, bboxes, new_image_ratio):
        r"""
        Removes padding from predicted bboxes for previously padded image

        Parameters
        ----------
        bboxes : list of lists
            list of bboxes that have to be recalculated to remove implied padding
        new_image_ratio : float
            Ratio of padded image size to the original image size

        Returns
        -------
        new_bboxes: list
            bboxes with transformed coordinates
        """
        new_bboxes = []
        c_x = 0.5
        c_y = 0.5

        self._log().debug("PREDICTED BBOXES: {}".format(bboxes))
        self._log().debug("new_image_ratio: {}".format(new_image_ratio))

        for bbox in bboxes:
            # 1. corner coords -> center coords
            cb_x1 = bbox[0] - c_x
            cb_y1 = bbox[1] - c_y
            cb_x2 = bbox[2] - c_x
            cb_y2 = bbox[3] - c_y

            # 2. center coords * new_image_ratio
            r_cb_x1 = cb_x1 * new_image_ratio
            r_cb_y1 = cb_y1 * new_image_ratio
            r_cb_x2 = cb_x2 * new_image_ratio
            r_cb_y2 = cb_y2 * new_image_ratio

            # 3. center coords -> corner coords
            x1 = r_cb_x1 + c_x
            y1 = r_cb_y1 + c_y
            x2 = r_cb_x2 + c_x
            y2 = r_cb_y2 + c_y

            x1 = np.clip(x1, 0.0, 1.0)
            y1 = np.clip(y1, 0.0, 1.0)
            x2 = np.clip(x2, 0.0, 1.0)
            y2 = np.clip(y2, 0.0, 1.0)

            new_bbox = [x1, y1, x2, y2]
            new_bboxes.append(new_bbox)

        self._log().debug("DEPAD BBOXES: {}".format(new_bboxes))

        return new_bboxes

    def _merge_tf_output(self, docling_output, pdf_cells):
        tf_output = []
        tf_cells_map = {}
        max_row_idx = 0

        for docling_item in docling_output:
            r_idx = str(docling_item["start_row_offset_idx"])
            c_idx = str(docling_item["start_col_offset_idx"])
            cell_key = c_idx + "_" + r_idx
            if cell_key in tf_cells_map:
                for pdf_cell in pdf_cells:
                    if pdf_cell["id"] == docling_item["cell_id"]:
                        text_cell_bbox = {
                            "b": pdf_cell["bbox"][3],
                            "l": pdf_cell["bbox"][0],
                            "r": pdf_cell["bbox"][2],
                            "t": pdf_cell["bbox"][1],
                            "token": pdf_cell["text"],
                        }
                        tf_cells_map[cell_key]["text_cell_bboxes"].append(
                            text_cell_bbox
                        )
            else:
                tf_cells_map[cell_key] = {
                    "bbox": docling_item["bbox"],
                    "row_span": docling_item["row_span"],
                    "col_span": docling_item["col_span"],
                    "start_row_offset_idx": docling_item["start_row_offset_idx"],
                    "end_row_offset_idx": docling_item["end_row_offset_idx"],
                    "start_col_offset_idx": docling_item["start_col_offset_idx"],
                    "end_col_offset_idx": docling_item["end_col_offset_idx"],
                    "indentation_level": docling_item["indentation_level"],
                    "text_cell_bboxes": [],
                    "column_header": docling_item["column_header"],
                    "row_header": docling_item["row_header"],
                    "row_section": docling_item["row_section"],
                }

                if docling_item["start_row_offset_idx"] > max_row_idx:
                    max_row_idx = docling_item["start_row_offset_idx"]

                for pdf_cell in pdf_cells:
                    if pdf_cell["id"] == docling_item["cell_id"]:
                        text_cell_bbox = {
                            "b": pdf_cell["bbox"][3],
                            "l": pdf_cell["bbox"][0],
                            "r": pdf_cell["bbox"][2],
                            "t": pdf_cell["bbox"][1],
                            "token": pdf_cell["text"],
                        }
                        tf_cells_map[cell_key]["text_cell_bboxes"].append(
                            text_cell_bbox
                        )

        for k in tf_cells_map:
            tf_output.append(tf_cells_map[k])
        return tf_output

    def resize_img(self, image, width=None, height=None, inter=cv2.INTER_AREA):
        # initialize the dimensions of the image to be resized and
        # grab the image size
        dim = None
        (h, w) = image.shape[:2]
        sf = 1.0
        # if both the width and height are None, then return the
        # original image
        if width is None and height is None:
            return image, sf
        # check to see if the width is None
        if width is None:
            # calculate the ratio of the height and construct the
            # dimensions
            r = height / float(h)
            sf = r
            dim = (int(w * r), height)
        # otherwise, the height is None
        else:
            # calculate the ratio of the width and construct the
            # dimensions
            r = width / float(w)
            sf = r
            dim = (width, int(h * r))
        # resize the image
        # TODO(Nikos): Try to remove cv2 dependency
        resized = cv2.resize(image, dim, interpolation=inter)
        # return the resized image
        return resized, sf

    def multi_table_predict(
        self,
        iocr_page,
        table_bboxes,
        do_matching=True,
        correct_overlapping_cells=False,
        sort_row_col_indexes=True,
    ):
        multi_tf_output = []
        page_image = iocr_page["image"]

        # Prevent large image submission, by resizing input
        page_image_resized, scale_factor = self.resize_img(page_image, height=1024)

        for table_bbox in table_bboxes:
            # Downscale table bounding box to the size of new image
            table_bbox[0] = table_bbox[0] * scale_factor
            table_bbox[1] = table_bbox[1] * scale_factor
            table_bbox[2] = table_bbox[2] * scale_factor
            table_bbox[3] = table_bbox[3] * scale_factor

            table_image = page_image_resized[
                round(table_bbox[1]) : round(table_bbox[3]),
                round(table_bbox[0]) : round(table_bbox[2]),
            ]
            # table_image = page_image
            # Predict
            if do_matching:
                tf_responses, predict_details = self.predict(
                    iocr_page,
                    table_bbox,
                    table_image,
                    scale_factor,
                    None,
                    correct_overlapping_cells,
                )
            else:
                tf_responses, predict_details = self.predict_dummy(
                    iocr_page, table_bbox, table_image, scale_factor, None
                )

            # ======================================================================================
            # PROCESS PREDICTED RESULTS, TO TURN PREDICTED COL/ROW IDs into Indexes
            # Indexes should be in increasing order, without gaps

            if sort_row_col_indexes:
                # Fix col/row indexes
                # Arranges all col/row indexes sequentially without gaps using input IDs

                indexing_start_cols = (
                    []
                )  # Index of original start col IDs (not indexes)
                indexing_end_cols = []  # Index of original end col IDs (not indexes)
                indexing_start_rows = (
                    []
                )  # Index of original start row IDs (not indexes)
                indexing_end_rows = []  # Index of original end row IDs (not indexes)

                # First, collect all possible predicted IDs, to be used as indexes
                # ID's returned by Tableformer are sequential, but might contain gaps
                for tf_response_cell in tf_responses:
                    start_col_offset_idx = tf_response_cell["start_col_offset_idx"]
                    end_col_offset_idx = tf_response_cell["end_col_offset_idx"]
                    start_row_offset_idx = tf_response_cell["start_row_offset_idx"]
                    end_row_offset_idx = tf_response_cell["end_row_offset_idx"]

                    # Collect all possible col/row IDs:
                    if start_col_offset_idx not in indexing_start_cols:
                        indexing_start_cols.append(start_col_offset_idx)
                    if end_col_offset_idx not in indexing_end_cols:
                        indexing_end_cols.append(end_col_offset_idx)
                    if start_row_offset_idx not in indexing_start_rows:
                        indexing_start_rows.append(start_row_offset_idx)
                    if end_row_offset_idx not in indexing_end_rows:
                        indexing_end_rows.append(end_row_offset_idx)

                indexing_start_cols.sort()
                indexing_end_cols.sort()
                indexing_start_rows.sort()
                indexing_end_rows.sort()

                # After this - put actual indexes of IDs back into predicted structure...
                for tf_response_cell in tf_responses:
                    tf_response_cell["start_col_offset_idx"] = (
                        indexing_start_cols.index(
                            tf_response_cell["start_col_offset_idx"]
                        )
                    )
                    tf_response_cell["end_col_offset_idx"] = (
                        tf_response_cell["start_col_offset_idx"]
                        + tf_response_cell["col_span"]
                    )
                    tf_response_cell["start_row_offset_idx"] = (
                        indexing_start_rows.index(
                            tf_response_cell["start_row_offset_idx"]
                        )
                    )
                    tf_response_cell["end_row_offset_idx"] = (
                        tf_response_cell["start_row_offset_idx"]
                        + tf_response_cell["row_span"]
                    )
                # Counting matched cols/rows from actual indexes (and not ids)
                predict_details["num_cols"] = len(indexing_end_cols)
                predict_details["num_rows"] = len(indexing_end_rows)
            else:
                otsl_seq = predict_details["prediction"]["rs_seq"]
                predict_details["num_cols"] = otsl_seq.index("nl")
                predict_details["num_rows"] = otsl_seq.count("nl")

            # Put results into multi_tf_output
            multi_tf_output.append(
                {"tf_responses": tf_responses, "predict_details": predict_details}
            )
            # Upscale table bounding box back, for visualization purposes
            table_bbox[0] = table_bbox[0] / scale_factor
            table_bbox[1] = table_bbox[1] / scale_factor
            table_bbox[2] = table_bbox[2] / scale_factor
            table_bbox[3] = table_bbox[3] / scale_factor
        # Return grouped results of predictions
        return multi_tf_output

    def predict_dummy(
        self, iocr_page, table_bbox, table_image, scale_factor, eval_res_preds=None
    ):
        r"""
        Predict the table out of an image in memory

        Parameters
        ----------
        iocr_page : dict
            Docling provided table data
        eval_res_preds : dict
            Ready predictions provided by the evaluation results

        Returns
        -------
        docling_output : string
            json response formatted according to Docling api expectations

        matching_details : string
            json with details about the matching between the pdf cells and the table cells
        """
        AggProfiler().start_agg(self._prof)

        max_steps = self._config["predict"]["max_steps"]
        beam_size = self._config["predict"]["beam_size"]
        image_batch = self._prepare_image(table_image)
        # Make predictions
        prediction = {}

        with torch.no_grad():
            # Compute predictions
            if (
                eval_res_preds is not None
            ):  # Don't run the model, use the provided predictions
                prediction["bboxes"] = eval_res_preds["bboxes"]
                pred_tag_seq = eval_res_preds["tag_seq"]
            elif self._config["predict"]["bbox"]:
                pred_tag_seq, outputs_class, outputs_coord = self._model.predict(
                    image_batch, max_steps, beam_size
                )

                if outputs_coord is not None:
                    if len(outputs_coord) == 0:
                        prediction["bboxes"] = []
                    else:
                        bbox_pred = u.box_cxcywh_to_xyxy(outputs_coord)
                        prediction["bboxes"] = bbox_pred.tolist()
                else:
                    prediction["bboxes"] = []

                if outputs_class is not None:
                    if len(outputs_class) == 0:
                        prediction["classes"] = []
                    else:
                        result_class = torch.argmax(outputs_class, dim=1)
                        prediction["classes"] = result_class.tolist()
                else:
                    prediction["classes"] = []
                if self._remove_padding:
                    pred_tag_seq, _ = u.remove_padding(pred_tag_seq)
            else:
                pred_tag_seq, _, _ = self._model.predict(
                    image_batch, max_steps, beam_size
                )
                # Check if padding should be removed
                if self._remove_padding:
                    pred_tag_seq, _ = u.remove_padding(pred_tag_seq)

            prediction["tag_seq"] = pred_tag_seq
            prediction["rs_seq"] = self._get_html_tags(pred_tag_seq)
            prediction["html_seq"] = otsl_to_html(prediction["rs_seq"], False)
        # Remove implied padding from bbox predictions,
        # that we added on image pre-processing stage
        self._log().debug("----- rs_seq -----")
        self._log().debug(prediction["rs_seq"])
        self._log().debug(len(prediction["rs_seq"]))
        otsl_sqr_chk(prediction["rs_seq"], False)

        # Check that bboxes are in sync with predicted tags
        sync, corrected_bboxes = self._check_bbox_sync(prediction)
        if not sync:
            prediction["bboxes"] = corrected_bboxes

        # Match the cells
        matching_details = {
            "table_cells": [],
            "matches": {},
            "pdf_cells": [],
            "prediction_bboxes_page": [],
        }

        # Table bbox upscaling will scale predicted bboxes too within cell matcher
        scaled_table_bbox = [
            table_bbox[0] / scale_factor,
            table_bbox[1] / scale_factor,
            table_bbox[2] / scale_factor,
            table_bbox[3] / scale_factor,
        ]

        if len(prediction["bboxes"]) > 0:
            matching_details = self._cell_matcher.match_cells_dummy(
                iocr_page, scaled_table_bbox, prediction
            )
            # Generate the expected Docling responses
            AggProfiler().begin("generate_docling_response", self._prof)
            docling_output = self._generate_tf_response_dummy(
                matching_details["table_cells"]
            )

            AggProfiler().end("generate_docling_response", self._prof)
            # Add the docling_output sorted by cell_id into the matching_details
            docling_output.sort(key=lambda item: item["cell_id"])
            matching_details["docling_responses"] = docling_output
            # Merge docling_output and pdf_cells into one TF output,
            # with deduplicated table cells
            # tf_output = self._merge_tf_output_dummy(docling_output)
            tf_output = docling_output

        return tf_output, matching_details

    def predict(
        self,
        iocr_page,
        table_bbox,
        table_image,
        scale_factor,
        eval_res_preds=None,
        correct_overlapping_cells=False,
    ):
        r"""
        Predict the table out of an image in memory

        Parameters
        ----------
        iocr_page : dict
            Docling provided table data
        eval_res_preds : dict
            Ready predictions provided by the evaluation results
        correct_overlapping_cells : boolean
            Enables or disables last post-processing step, that fixes cell bboxes to remove overlap

        Returns
        -------
        docling_output : string
            json response formatted according to Docling api expectations

        matching_details : string
            json with details about the matching between the pdf cells and the table cells
        """
        AggProfiler().start_agg(self._prof)

        max_steps = self._config["predict"]["max_steps"]
        beam_size = self._config["predict"]["beam_size"]
        image_batch = self._prepare_image(table_image)
        # Make predictions
        prediction = {}

        with torch.no_grad():
            # Compute predictions
            if (
                eval_res_preds is not None
            ):  # Don't run the model, use the provided predictions
                prediction["bboxes"] = eval_res_preds["bboxes"]
                pred_tag_seq = eval_res_preds["tag_seq"]
            elif self._config["predict"]["bbox"]:
                pred_tag_seq, outputs_class, outputs_coord = self._model.predict(
                    image_batch, max_steps, beam_size
                )

                if outputs_coord is not None:
                    if len(outputs_coord) == 0:
                        prediction["bboxes"] = []
                    else:
                        bbox_pred = u.box_cxcywh_to_xyxy(outputs_coord)
                        prediction["bboxes"] = bbox_pred.tolist()
                else:
                    prediction["bboxes"] = []

                if outputs_class is not None:
                    if len(outputs_class) == 0:
                        prediction["classes"] = []
                    else:
                        result_class = torch.argmax(outputs_class, dim=1)
                        prediction["classes"] = result_class.tolist()
                else:
                    prediction["classes"] = []
                if self._remove_padding:
                    pred_tag_seq, _ = u.remove_padding(pred_tag_seq)
            else:
                pred_tag_seq, _, _ = self._model.predict(
                    image_batch, max_steps, beam_size
                )
                # Check if padding should be removed
                if self._remove_padding:
                    pred_tag_seq, _ = u.remove_padding(pred_tag_seq)

            prediction["tag_seq"] = pred_tag_seq
            prediction["rs_seq"] = self._get_html_tags(pred_tag_seq)
            prediction["html_seq"] = otsl_to_html(prediction["rs_seq"], False)
        # Remove implied padding from bbox predictions,
        # that we added on image pre-processing stage
        self._log().debug("----- rs_seq -----")
        self._log().debug(prediction["rs_seq"])
        self._log().debug(len(prediction["rs_seq"]))
        otsl_sqr_chk(prediction["rs_seq"], False)

        sync, corrected_bboxes = self._check_bbox_sync(prediction)
        if not sync:
            prediction["bboxes"] = corrected_bboxes

        # Match the cells
        matching_details = {
            "table_cells": [],
            "matches": {},
            "pdf_cells": [],
            "prediction_bboxes_page": [],
        }

        # Table bbox upscaling will scale predicted bboxes too within cell matcher
        scaled_table_bbox = [
            table_bbox[0] / scale_factor,
            table_bbox[1] / scale_factor,
            table_bbox[2] / scale_factor,
            table_bbox[3] / scale_factor,
        ]

        if len(prediction["bboxes"]) > 0:
            matching_details = self._cell_matcher.match_cells(
                iocr_page, scaled_table_bbox, prediction
            )
        # Post-processing
        if len(prediction["bboxes"]) > 0:
            if (
                len(iocr_page["tokens"]) > 0
            ):  # There are at least some pdf cells to match with
                if self.enable_post_process:
                    AggProfiler().begin("post_process", self._prof)
                    matching_details = self._post_processor.process(
                        matching_details, correct_overlapping_cells
                    )
                    AggProfiler().end("post_process", self._prof)

        # Generate the expected Docling responses
        AggProfiler().begin("generate_docling_response", self._prof)
        docling_output = self._generate_tf_response(
            matching_details["table_cells"],
            matching_details["matches"],
        )

        AggProfiler().end("generate_docling_response", self._prof)
        # Add the docling_output sorted by cell_id into the matching_details
        docling_output.sort(key=lambda item: item["cell_id"])
        matching_details["docling_responses"] = docling_output

        # Merge docling_output and pdf_cells into one TF output,
        # with deduplicated table cells
        tf_output = self._merge_tf_output(docling_output, matching_details["pdf_cells"])

        return tf_output, matching_details

    def _generate_tf_response_dummy(self, table_cells):
        tf_cell_list = []

        for table_cell in table_cells:
            colspan_val = 1
            if "colspan_val" in table_cell:
                colspan_val = table_cell["colspan_val"]
            rowspan_val = 1
            if "rowspan_val" in table_cell:
                rowspan_val = table_cell["rowspan_val"]

            column_header = False
            if table_cell["label"] == "ched":
                column_header = True

            row_header = False
            if table_cell["label"] == "rhed":
                row_header = True

            row_section = False
            if table_cell["label"] == "srow":
                row_section = True

            row_id = table_cell["row_id"]
            column_id = table_cell["column_id"]

            cell_bbox = {
                "b": table_cell["bbox"][3],
                "l": table_cell["bbox"][0],
                "r": table_cell["bbox"][2],
                "t": table_cell["bbox"][1],
                "token": "",
            }

            tf_cell = {
                "cell_id": table_cell["cell_id"],
                "bbox": cell_bbox,  # b,l,r,t,token
                "row_span": rowspan_val,
                "col_span": colspan_val,
                "start_row_offset_idx": row_id,
                "end_row_offset_idx": row_id + rowspan_val,
                "start_col_offset_idx": column_id,
                "end_col_offset_idx": column_id + colspan_val,
                "indentation_level": 0,
                # No text cell bboxes, because no matching was done
                "text_cell_bboxes": [],
                "column_header": column_header,
                "row_header": row_header,
                "row_section": row_section,
            }
            tf_cell_list.append(tf_cell)
        return tf_cell_list

    def _generate_tf_response(self, table_cells, matches):
        r"""
        Convert the matching details to the expected output for Docling

        Parameters
        ----------
        table_cells : list of dict
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id",
                                                  "bbox", "label", "class"
        matches : dictionary of lists of table_cells
            A dictionary which is indexed by the pdf_cell_id as key and the value is a list
            of the table_cells that fall inside that pdf cell

        Returns
        -------
        docling_output : string
            json response formatted according to Docling api expectations
        """

        # format output to look similar to tests/examples/tf_gte_output_2.json
        tf_cell_list = []
        for pdf_cell_id, pdf_cell_matches in matches.items():
            tf_cell = {
                "bbox": {},  # b,l,r,t,token
                "row_span": 1,
                "col_span": 1,
                "start_row_offset_idx": -1,
                "end_row_offset_idx": -1,
                "start_col_offset_idx": -1,
                "end_col_offset_idx": -1,
                "indentation_level": 0,
                # return text cell bboxes additionally to the matched index
                "text_cell_bboxes": [{}],  # b,l,r,t,token
                "column_header": False,
                "row_header": False,
                "row_section": False,
            }
            tf_cell["cell_id"] = int(pdf_cell_id)

            row_ids = set()
            column_ids = set()
            labels = set()

            for match in pdf_cell_matches:
                tm = match["table_cell_id"]
                tcl = list(
                    filter(lambda table_cell: table_cell["cell_id"] == tm, table_cells)
                )
                if len(tcl) > 0:
                    table_cell = tcl[0]
                    row_ids.add(table_cell["row_id"])
                    column_ids.add(table_cell["column_id"])
                    labels.add(table_cell["label"])

                    if table_cell["label"] is not None:
                        if table_cell["label"] in ["ched"]:
                            tf_cell["column_header"] = True
                        if table_cell["label"] in ["rhed"]:
                            tf_cell["row_header"] = True
                        if table_cell["label"] in ["srow"]:
                            tf_cell["row_section"] = True

                    tf_cell["start_col_offset_idx"] = table_cell["column_id"]
                    tf_cell["end_col_offset_idx"] = table_cell["column_id"] + 1
                    tf_cell["start_row_offset_idx"] = table_cell["row_id"]
                    tf_cell["end_row_offset_idx"] = table_cell["row_id"] + 1

                    if "colspan_val" in table_cell:
                        tf_cell["col_span"] = table_cell["colspan_val"]
                        tf_cell["start_col_offset_idx"] = table_cell["column_id"]
                        off_idx = table_cell["column_id"] + tf_cell["col_span"]
                        tf_cell["end_col_offset_idx"] = off_idx
                    if "rowspan_val" in table_cell:
                        tf_cell["row_span"] = table_cell["rowspan_val"]
                        tf_cell["start_row_offset_idx"] = table_cell["row_id"]
                        tf_cell["end_row_offset_idx"] = (
                            table_cell["row_id"] + tf_cell["row_span"]
                        )
                    if "bbox" in table_cell:
                        table_match_bbox = table_cell["bbox"]
                        tf_bbox = {
                            "b": table_match_bbox[3],
                            "l": table_match_bbox[0],
                            "r": table_match_bbox[2],
                            "t": table_match_bbox[1],
                        }
                        tf_cell["bbox"] = tf_bbox

            tf_cell["row_ids"] = list(row_ids)
            tf_cell["column_ids"] = list(column_ids)
            tf_cell["label"] = "None"
            l_labels = list(labels)
            if len(l_labels) > 0:
                tf_cell["label"] = l_labels[0]
            tf_cell_list.append(tf_cell)
        return tf_cell_list

    def _prepare_image(self, mat_image):
        r"""
        Rescale the image and prepare a batch of 1 with the image as as tensor

        Parameters
        ----------
        mat_image: cv2.Mat
            The image as an openCV Mat object

        Returns
        -------
        tensor (batch_size, image_channels, resized_image, resized_image)
        """
        normalize = T.Normalize(
            mean=self._config["dataset"]["image_normalization"]["mean"],
            std=self._config["dataset"]["image_normalization"]["std"],
        )
        resized_size = self._config["dataset"]["resized_image"]
        resize = T.Resize([resized_size, resized_size])

        img, _ = normalize(mat_image, None)
        img, _ = resize(img, None)

        img = img.transpose(2, 1, 0)  # (channels, width, height)
        img = torch.FloatTensor(img / 255.0)
        image_batch = img.unsqueeze(dim=0)
        image_batch = image_batch.to(device=self._device)
        return image_batch

    def _get_html_tags(self, seq):
        r"""
        Convert indices to actual html tags

        """
        # Map the tag indices back to actual tags (without start, end)
        html_tags = [self._rev_word_map[ind] for ind in seq[1:-1]]

        return html_tags



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import os
import json
import glob
import copy

import logging

from pathlib import Path

import numpy as np
import pytest
from PIL import Image

from datasets import load_dataset

from typing import List, Dict
import random

from docling_ibm_models.reading_order.reading_order_rb import PageElement, ReadingOrderPredictor

from docling_core.types.doc.document import DoclingDocument, DocItem, RefItem, TextItem, ContentLayer

# Configure logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)

def rank_array(arr):
    """Compute ranks, resolving ties by averaging."""
    sorted_indices = sorted(range(len(arr)), key=lambda i: arr[i])  # Sort indices
    ranks = [0] * len(arr)  # Initialize ranks
    
    i = 0
    while i < len(arr):
        start = i
        while i + 1 < len(arr) and arr[sorted_indices[i]] == arr[sorted_indices[i + 1]]:
            i += 1  # Handle ties
        avg_rank = sum(range(start + 1, i + 2)) / (i - start + 1)  # Average rank for ties
        for j in range(start, i + 1):
            ranks[sorted_indices[j]] = avg_rank
        i += 1
    return ranks

def spearman_rank_correlation(arr1, arr2):
    assert len(arr1) == len(arr2), "Arrays must have the same length"
    
    # Compute ranks
    rank1 = rank_array(arr1)
    rank2 = rank_array(arr2)

    # Compute rank differences and apply formula
    d = [rank1[i] - rank2[i] for i in range(len(arr1))]
    d_squared_sum = sum(d_i ** 2 for d_i in d)

    n = len(arr1)
    if n > 1:
        rho = 1 - (6 * d_squared_sum) / (n * (n**2 - 1))
    else:
        rho = 0
    return rho
            
def test_readingorder():

    ro_scores, caption_scores, footnote_scores = [], [], []
    
    # Init the reading-order model
    romodel = ReadingOrderPredictor()

    ds = load_dataset("ds4sd/docling-dpbench")
    for row in ds["test"]:
        true_doc = DoclingDocument.model_validate_json(row["GroundTruthDocument"])
        
        true_elements: List[PageElement] = []
        pred_elements: List[PageElement] = []

        to_ref: Dict[int, str] = {}
        from_ref: Dict[str, int] = {}
        
        for item, level in true_doc.iterate_items(included_content_layers={ContentLayer.BODY, ContentLayer.FURNITURE}):
            if isinstance(item, DocItem):
                for prov in item.prov:

                    page_height = true_doc.pages[prov.page_no].size.height
                    bbox = prov.bbox.to_bottom_left_origin(page_height=page_height)

                    text = ""
                    if isinstance(item, TextItem):
                        text = item.text
                    
                    true_elements.append(
                        PageElement(
                            cid=len(true_elements),
                            ref=item.get_ref(),
                            text = text,
                            page_no=prov.page_no,
                            page_size = true_doc.pages[prov.page_no].size,
                            label=item.label,
                            l = bbox.l,
                            r = bbox.r,
                            b = bbox.b,
                            t = bbox.t,
                            coord_origin = bbox.coord_origin
                        )
                    )

                    to_ref[true_elements[-1].cid] = item.get_ref().cref 
                    from_ref[item.get_ref().cref] = true_elements[-1].cid 
                    
        rand_elements = copy.deepcopy(true_elements)
        random.shuffle(rand_elements)

        """
        print(f"reading {os.path.basename(filename)}")        
        for true_elem, rand_elem in zip(true_elements, rand_elements):
            print("true: ", str(true_elem), ", rand: ", str(rand_elem))
        """
        
        pred_elements = romodel.predict_reading_order(page_elements=rand_elements)
        #pred_elements = romodel.predict_page(page_elements=rand_elements)    

        assert len(pred_elements)==len(true_elements), f"{len(pred_elements)}!={len(true_elements)}"

        true_cids, pred_cids = [], []
        for true_elem, pred_elem, rand_elem in zip(true_elements,
                                                   pred_elements,
                                                   rand_elements):
            true_cids.append(true_elem.cid)
            pred_cids.append(pred_elem.cid)

        score = spearman_rank_correlation(true_cids, pred_cids)
        ro_scores.append(score)
        
        filename = row["document_id"]

        if score == 0:
            continue
        # Identify special cases ...
        if filename in ["doc_906d54a21ef3c7bfac03f4bb613b0c79ef32fdf81b362450c79e98a96f88708a_page_000001.png",
                        "doc_2cd17a32ee330a239e19c915738df0c27e8ec3635a60a7e16e2a0cf3868d4af3_page_000001.png",
                        "doc_bcb3dafc35b5e7476fd1b9cd6eccf5eeef936cd5b13ad846a4943f1e7797f4e9_page_000001.png",
                        "doc_a0edae1fa147c7bb78ebc493743a68ba4372b5ead31f2a2b146c35119462379e_page_000001.png",
                        "doc_94ba5468fcb6277721947697048846dc0d0551296be3b45f5918ab857d21dcc7_page_000001.png",
                        "doc_cbb4a13ffd01d9f777fdb939451d6a21cea1b869ee50d79581451e3601df9ec8_page_000001.png"]:
            # print(f"{os.path.basename(filename)}: {score}")
            assert score>=0.60, f"reading-order score={score}>0.60"            
        else:
            assert score>=0.90, f"reading-order score={score}>0.90 for {filename}"


        true_to_captions: Dict[int, List[int]] = {}
        true_to_footnotes: Dict[int, List[int]] = {}

        total_caption_links = 0
        total_footnote_links = 0
        
        for table in true_doc.tables:
            table_cid = from_ref[table.get_ref().cref]

            true_to_captions[table_cid] = []  
            for caption in table.captions:
                caption_cid = from_ref[caption.get_ref().cref]
                true_to_captions[table_cid].append(caption_cid) 

                total_caption_links += 1

            true_to_footnotes[table_cid] = []  
            for footnote in table.footnotes:
                footnote_cid = from_ref[footnote.get_ref().cref]
                true_to_footnotes[table_cid].append(footnote_cid) 

                total_footnote_links += 1
                
        for picture in true_doc.pictures:
            picture_cid = from_ref[picture.get_ref().cref]

            true_to_captions[picture_cid] = []  
            for caption in picture.captions:
                caption_cid = from_ref[caption.get_ref().cref]
                true_to_captions[picture_cid].append(caption_cid)                 

                total_caption_links += 1

            true_to_footnotes[picture_cid] = []  
            for footnote in picture.footnotes:
                footnote_cid = from_ref[footnote.get_ref().cref]
                true_to_footnotes[picture_cid].append(footnote_cid)                 

                total_footnote_links += 1
                
        if total_caption_links>0:
            #print(" *********** ")
            
            pred_to_captions = romodel.predict_to_captions(sorted_elements=pred_elements)

            """
            for key,val in pred_to_captions.items():
                print(f"pred {key}: {val}")
            """
            
            score, total = 0.0, 0.0
            for key,val in true_to_captions.items():
                # print(f"true {key}: {val}")            
                
                total += 1.0
                if key in pred_to_captions and pred_to_captions[key]==val:
                    score += 1.0

            # print(f"to_captions: {score/total}")
            caption_scores.append(score/total)

        if total_footnote_links>0:
            # print(" *********** ")
            
            pred_to_footnotes = romodel.predict_to_footnotes(sorted_elements=pred_elements)

            """
            for key,val in pred_to_footnotes.items():
                print(f"pred {key}: {val}")
            """
            
            score, total = 0.0, 0.0
            for key,val in true_to_footnotes.items():
                # print(f"true {key}: {val}")            
                
                total += 1.0
                if key in pred_to_footnotes and pred_to_footnotes[key]==val:
                    score += 1.0

            # print(f"to_footnotes: {score/total}")
            footnote_scores.append(score/total)

        pred_merges = romodel.predict_merges(sorted_elements=pred_elements)
        # print("merges: ", pred_merges)
                    
                    
    mean_ro_score = sum(ro_scores)/len(ro_scores)
    mean_cp_score = sum(caption_scores)/len(caption_scores)
    mean_ft_score = sum(footnote_scores)/len(footnote_scores)

    assert mean_ro_score>0.95, "mean_ro_score>0.95"
    assert mean_cp_score>0.85, "mean_cp_score>0.85"
    assert mean_ft_score>0.90, "mean_ft_score>0.90"
    
    print("\n  score(reading): ", mean_ro_score)
    print("  score(caption): ", mean_cp_score)
    print("score(footnotes): ", mean_ft_score)


"""    
def test_readingorder_multipage():

    filename = Path("<json with page-elements>")
    
    # Init the reading-order model
    romodel = ReadingOrderPredictor()

    true_elements: List[PageElement] = []
    pred_elements: List[PageElement] = []
    
    with open(filename, "r") as fr:
        data = json.load(fr)
        true_elements = [PageElement.model_validate(item) for item in data]

    pred_elements = romodel.predict_reading_order(page_elements=true_elements)
    for true_elem, pred_elem in zip(true_elements, pred_elements):
        print("true: ", str(true_elem), ", pred: ", str(pred_elem))
"""
    



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import glob
import json
import os
from pathlib import Path

import torch
import pytest
import cv2
from PIL import Image, ImageDraw
from huggingface_hub import snapshot_download

from docling_ibm_models.tableformer.utils.app_profiler import AggProfiler
import docling_ibm_models.tableformer.data_management.tf_predictor as tf_predictor
from docling_ibm_models.tableformer.data_management.tf_predictor import \
    TFPredictor

"""
- Implements TF predictor to accept the input format from IOCR, e.g.
  "./tests/test_data/samples/tf_table_example_0.json" (trivial table crop)
- Shape the output format like GTE would: e.g.
  "tests/test_data/samples/tf_gte_output_2.json" (Note: full form image)
"""

docling_api_data = {
    "table_jsons": [
        "./tests/test_data/samples/ADS.2007.page_123.png_iocr.parse_format.json",
        "./tests/test_data/samples/PHM.2013.page_30.png_iocr.parse_format.json",
        "./tests/test_data/samples/empty_iocr.png.json"
    ],
    "png_images": [
        "./tests/test_data/samples/ADS.2007.page_123.png",
        "./tests/test_data/samples/PHM.2013.page_30.png",
        "./tests/test_data/samples/empty_iocr.png"
    ],
    "table_bboxes": [
        [[178, 748, 1061, 976], [177, 1163, 1062, 1329]],
        [[100, 186, 1135, 525]],
        [[178, 748, 1061, 976], [177, 1163, 1062, 1329]]
    ],
}

# The config is missing the keys: "model.save_dir"
test_config = {
    "dataset": {
        "type": "TF_prepared",
        "name": "TF",
        "raw_data_dir": "./tests/test_data/model_artifacts/",
        "load_cells": True,
        "bbox_format": "5plet",
        "resized_image": 448,
        "keep_AR": False,
        "up_scaling_enabled": True,
        "down_scaling_enabled": True,
        "padding_mode": "null",
        "padding_color": [0, 0, 0],
        "image_normalization": {
            "state": True,
            "mean": [0.94247851, 0.94254675, 0.94292611],
            "std": [0.17910956, 0.17940403, 0.17931663],
        },
        "color_jitter": True,
        "rand_crop": True,
        "rand_pad": True,
        "image_grayscale": False,
    },
    "model": {
        "type": "TableModel04_rs",
        "name": "14_128_256_4_true",
        # "save_dir": "./tests/test_data/model_artifacts/",
        "backbone": "resnet18",
        "enc_image_size": 28,
        "tag_embed_dim": 16,
        "hidden_dim": 512,
        "tag_decoder_dim": 512,
        "bbox_embed_dim": 256,
        "tag_attention_dim": 256,
        "bbox_attention_dim": 512,
        "enc_layers": 4,  # 6
        "dec_layers": 2,  # 6
        "nheads": 8,
        "dropout": 0.1,
        "bbox_classes": 2,
    },
    "train": {
        "save_periodicity": 1,
        "disable_cuda": False,
        "epochs": 23,
        "batch_size": 50,
        "clip_gradient": 0.1,
        "clip_max_norm": 0.1,
        "bbox": True,
        "validation": False,
    },
    "predict": {
        "max_steps": 1024,
        "beam_size": 5,
        "bbox": True,
        "predict_dir": "./tests/test_data/samples",
        "pdf_cell_iou_thres": 0.05,
        "padding": False,
        "padding_size": 50,
        "disable_post_process": False,
        "profiling": True
    },
    "dataset_wordmap": {
        "word_map_tag": {
            "<pad>": 0,
            "<unk>": 1,
            "<start>": 2,
            "<end>": 3,
            "ecel": 4,
            "fcel": 5,
            "lcel": 6,
            "ucel": 7,
            "xcel": 8,
            "nl": 9,
            "ched": 10,
            "rhed": 11,
            "srow": 12,
        },
        "word_map_cell": {
            " ": 13,
            "!": 179,
            '"': 126,
            "#": 101,
            "$": 119,
            "%": 18,
            "&": 114,
            "'": 108,
            "(": 29,
            ")": 32,
            "*": 26,
            "+": 97,
            ",": 71,
            "-": 63,
            ".": 34,
            "/": 66,
            "0": 33,
            "1": 36,
            "2": 43,
            "3": 41,
            "4": 45,
            "5": 17,
            "6": 37,
            "7": 35,
            "8": 40,
            "9": 16,
            ":": 88,
            ";": 92,
            "<": 73,
            "</b>": 9,
            "</i>": 23,
            "</overline>": 219,
            "</strike>": 233,
            "</sub>": 94,
            "</sup>": 77,
            "</underline>": 151,
            "<b>": 1,
            "<end>": 280,
            "<i>": 21,
            "<overline>": 218,
            "<pad>": 0,
            "<start>": 279,
            "<strike>": 232,
            "<sub>": 93,
            "<sup>": 75,
            "<underline>": 150,
            "<unk>": 278,
            "=": 99,
            ">": 39,
            "?": 96,
            "@": 125,
            "A": 27,
            "B": 86,
            "C": 19,
            "D": 57,
            "E": 64,
            "F": 47,
            "G": 44,
            "H": 10,
            "I": 20,
            "J": 80,
            "K": 81,
            "L": 52,
            "M": 46,
            "N": 69,
            "O": 65,
            "P": 62,
            "Q": 59,
            "R": 60,
            "S": 58,
            "T": 48,
            "U": 55,
            "V": 2,
            "W": 83,
            "X": 104,
            "Y": 89,
            "Z": 113,
            "[": 70,
            "\\": 165,
            "]": 72,
            "^": 132,
            "_": 84,
            "`": 196,
            "a": 3,
            "b": 6,
            "c": 54,
            "d": 12,
            "e": 8,
            "f": 50,
            "g": 28,
            "h": 56,
            "i": 5,
            "j": 82,
            "k": 95,
            "l": 7,
            "m": 30,
            "n": 31,
            "o": 15,
            "p": 22,
            "q": 67,
            "r": 4,
            "s": 51,
            "t": 14,
            "u": 25,
            "v": 24,
            "w": 53,
            "x": 61,
            "y": 49,
            "z": 11,
            "{": 158,
            "|": 139,
            "}": 159,
            "~": 147,
            "\u00a2": 203,
            "\u00a3": 162,
            "\u00a4": 220,
            "\u00a5": 176,
            "\u00a7": 142,
            "\u00a9": 268,
            "\u00ab": 239,
            "\u00ad": 275,
            "\u00ae": 130,
            "\u00b0": 100,
            "\u00b1": 79,
            "\u00b6": 171,
            "\u00b7": 137,
            "\u00bb": 240,
            "\u00d7": 118,
            "\u00d8": 192,
            "\u00df": 197,
            "\u00e6": 261,
            "\u00f7": 225,
            "\u00f8": 163,
            "\u0131": 242,
            "\u0142": 267,
            "\u01c2": 211,
            "\u025b": 223,
            "\u02b9": 248,
            "\u02c2": 195,
            "\u02c3": 208,
            "\u02c6": 253,
            "\u0300": 209,
            "\u0301": 131,
            "\u0302": 138,
            "\u0303": 156,
            "\u0304": 152,
            "\u0306": 222,
            "\u0307": 247,
            "\u0308": 103,
            "\u030a": 102,
            "\u030c": 254,
            "\u0327": 155,
            "\u0328": 269,
            "\u0338": 170,
            "\u0391": 173,
            "\u0392": 169,
            "\u0393": 180,
            "\u0394": 85,
            "\u0398": 243,
            "\u0399": 271,
            "\u039b": 272,
            "\u03a0": 213,
            "\u03a3": 185,
            "\u03a6": 148,
            "\u03a7": 212,
            "\u03a8": 141,
            "\u03a9": 161,
            "\u03b1": 90,
            "\u03b2": 107,
            "\u03b3": 110,
            "\u03b4": 153,
            "\u03b5": 166,
            "\u03b6": 178,
            "\u03b7": 146,
            "\u03b8": 186,
            "\u03b9": 229,
            "\u03ba": 164,
            "\u03bb": 91,
            "\u03bc": 78,
            "\u03bd": 230,
            "\u03be": 244,
            "\u03c0": 127,
            "\u03c1": 149,
            "\u03c3": 116,
            "\u03c4": 198,
            "\u03c5": 189,
            "\u03c6": 140,
            "\u03c7": 124,
            "\u03c8": 216,
            "\u03c9": 167,
            "\u0410": 273,
            "\u0421": 194,
            "\u115f": 217,
            "\u200b": 265,
            "\u2010": 117,
            "\u2012": 135,
            "\u2013": 42,
            "\u2014": 106,
            "\u2015": 228,
            "\u2016": 259,
            "\u2018": 123,
            "\u2019": 121,
            "\u201c": 87,
            "\u201d": 115,
            "\u201e": 245,
            "\u2020": 109,
            "\u2021": 129,
            "\u2022": 128,
            "\u2028": 190,
            "\u2030": 154,
            "\u2032": 68,
            "\u203b": 224,
            "\u2044": 188,
            "\u204e": 199,
            "\u2061": 200,
            "\u20ac": 184,
            "\u2190": 202,
            "\u2191": 112,
            "\u2192": 120,
            "\u2193": 111,
            "\u2194": 183,
            "\u21d1": 266,
            "\u21d2": 264,
            "\u21d3": 255,
            "\u2205": 215,
            "\u2206": 175,
            "\u2208": 262,
            "\u2211": 160,
            "\u2212": 76,
            "\u2216": 206,
            "\u2217": 105,
            "\u2218": 246,
            "\u2219": 236,
            "\u221a": 187,
            "\u221e": 207,
            "\u2223": 260,
            "\u2225": 193,
            "\u2227": 182,
            "\u2229": 256,
            "\u222b": 258,
            "\u223c": 98,
            "\u2248": 210,
            "\u2264": 38,
            "\u2265": 74,
            "\u2266": 214,
            "\u2267": 181,
            "\u2295": 263,
            "\u22c5": 174,
            "\u22c6": 191,
            "\u22ee": 277,
            "\u22ef": 270,
            "\u2500": 205,
            "\u2551": 231,
            "\u25a0": 250,
            "\u25a1": 177,
            "\u25aa": 145,
            "\u25b2": 136,
            "\u25b3": 143,
            "\u25bc": 251,
            "\u25c6": 226,
            "\u25ca": 235,
            "\u25cb": 227,
            "\u25cf": 172,
            "\u25e6": 274,
            "\u2605": 204,
            "\u2606": 144,
            "\u2640": 133,
            "\u2642": 134,
            "\u2663": 252,
            "\u2666": 157,
            "\u266f": 221,
            "\u2713": 122,
            "\u2714": 249,
            "\u2717": 201,
            "\u2794": 168,
            "\u27a2": 276,
            "\u2a7d": 234,
            "\u2a7e": 241,
            "\u3008": 237,
            "\u3009": 238,
            "\ufeff": 257,
        },
    },
}

# ==================================================================================================
configs = [test_config]


@pytest.fixture(scope="module")
def init() -> list[dict]:
    r"""
    Initialize the testing environment
    """
    # Download models from HF
    download_path = snapshot_download(repo_id="ds4sd/docling-models", revision="v2.1.0")
    save_dir = os.path.join(download_path, "model_artifacts/tableformer/fast")

    # Add the missing config keys
    for config in configs:
        config["model"]["save_dir"] = save_dir
    return configs

def test_tf_predictor(init):
    r"""
    Test the TFPredictor
    """
    viz = True
    device = "cpu"
    num_threads = 2

    # Load the docling_api_data
    iocr_pages = []
    for table_json_fn, png_image_fn, table_bboxes_b in zip(
        docling_api_data["table_jsons"],
        docling_api_data["png_images"],
        docling_api_data["table_bboxes"],
    ):
        with open(table_json_fn, "r") as fp:
            iocr_page_raw = json.load(fp)
            iocr_page = iocr_page_raw["pages"][0]
        # TODO(Nikos): Try to remove the opencv dependency
        iocr_page["image"] = cv2.imread(png_image_fn)
        iocr_page["png_image_fn"] = png_image_fn
        iocr_page["table_bboxes"] = table_bboxes_b
        iocr_pages.append(iocr_page)

    # Loop over the test configs
    for test_config in init:
        # Check if the checkpoint file should be combined
        # assert (
        #     combine_checkpoint(test_config["model"]["save_dir"]) >= 0
        # ), "Model checkpoint is missing"

        # Loop over the iocr_pages
        predictor = TFPredictor(test_config, device=device, num_threads=num_threads)
        for iocr_page in iocr_pages:
            # Prepare "Predict" parameters
            # iw = iocr_page["width"]
            # ih = iocr_page["height"]
            # table_bboxes = [[0, 0, iw, ih]]  # just one table per page in our examples
            table_bboxes = iocr_page["table_bboxes"]

            # for t, table_bbox in enumerate(table_bboxes):
            print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")
            png_img_bfn = os.path.basename(iocr_page["png_image_fn"])
            print("Predicting image: {}".format(png_img_bfn))

            # Run prediction, post-processing, and cell matching
            # PARAMETERS:
            # iocr_page - json received from iocr, augmented with iocr_page["image"]
            # table_bboxes - list of detected bboxes on page: [[x1, y1, x2, y2], [...]...]
            # do_matching - Boolean, when True - will match with text cells provided,
            #               when False - returns original cell prediction BBOXes in the same format
            # OUTPUT:
            # List of dicts per table: [{"tf_responses":[...], "predict_details": {}}]

            multi_tf_output = predictor.multi_table_predict(
                iocr_page,
                table_bboxes, 
                do_matching=True,
                correct_overlapping_cells=False,
                sort_row_col_indexes=True
            )

            # Test output for validity, create visualizations...
            for t, tf_output in enumerate(multi_tf_output):
                tf_responses = tf_output["tf_responses"]
                predict_details = tf_output["predict_details"]
                assert tf_responses is not None, "Empty prediction response"
                assert isinstance(
                    tf_responses, list
                ), " Wrong response type. It should be a list"

                img = Image.open(iocr_page["png_image_fn"])
                img1 = ImageDraw.Draw(img)

                xt0 = table_bboxes[t][0]
                yt0 = table_bboxes[t][1]
                xt1 = max(xt0, table_bboxes[t][2])
                yt1 = max(yt0, table_bboxes[t][3])
                img1.rectangle(((xt0, yt0), (xt1, yt1)), outline="pink", width=5)

                if viz:
                    # Visualize original OCR words:
                    for iocr_word in iocr_page["tokens"]:
                        xi0 = iocr_word["bbox"]["l"]
                        yi0 = iocr_word["bbox"]["t"]
                        xi1 = max(xi0, iocr_word["bbox"]["r"])
                        yi1 = max(yi0, iocr_word["bbox"]["b"])
                        img1.rectangle(((xi0, yi0), (xi1, yi1)), outline="gray")
                    # Visualize original docling_ibm_models.tableformer predictions:
                    for predicted_bbox in predict_details["prediction_bboxes_page"]:
                        xp0 = predicted_bbox[0] - 1
                        yp0 = predicted_bbox[1] - 1
                        xp1 = max(xp0, predicted_bbox[2] + 1)
                        yp1 = max(yp0, predicted_bbox[3] + 1)
                        img1.rectangle(((xp0, yp0), (xp1, yp1)), outline="green")

                # Check the structure of the list items
                for i, response in enumerate(tf_responses):
                    assert (
                        "bbox" in response
                    ), "bbox field is missing from response: " + str(i)
                    assert (
                        "text_cell_bboxes" in response
                    ), "text_cell_bboxes is missing: " + str(i)
                    assert (
                        "row_span" in response
                    ), "row_span is missing from resp: " + str(i)
                    assert (
                        "col_span" in response
                    ), "col_span is missing from response: " + str(i)
                    # print("*********** column_header: {}".format(response["column_header"]))
                    if viz:
                        # Visualization:
                        for text_cell in response["text_cell_bboxes"]:
                            xc0 = text_cell["l"]
                            yc0 = text_cell["t"]
                            xc1 = max(xc0, text_cell["r"])
                            yc1 = max(yc0, text_cell["b"])
                            img1.rectangle(((xc0, yc0), (xc1, yc1)), outline="red")

                        x0 = response["bbox"]["l"] - 2
                        y0 = response["bbox"]["t"] - 2
                        x1 = max(x0, response["bbox"]["r"] + 2)
                        y1 = max(y0, response["bbox"]["b"] + 2)

                        if response["column_header"]:
                            img1.rectangle(
                                ((x0, y0), (x1, y1)), outline="blue", width=2
                            )
                        elif response["row_header"]:
                            img1.rectangle(
                                ((x0, y0), (x1, y1)), outline="magenta", width=2
                            )
                        elif response["row_section"]:
                            img1.rectangle(
                                ((x0, y0), (x1, y1)), outline="brown", width=2
                            )
                        else:
                            img1.rectangle(
                                ((x0, y0), (x1, y1)), outline="black", width=1
                            )
                if viz:
                    viz_root = "./tests/test_data/viz/"
                    Path(viz_root).mkdir(parents=True, exist_ok=True)
                    png_img_bfn1 = png_img_bfn.replace(".png", "." + str(t) + ".png")
                    viz_fn = os.path.join(viz_root, png_img_bfn1)
                    img.save(viz_fn)

    # Get profiling data
    profiling_data = AggProfiler().get_data()
    print("Profiling data:")
    print(json.dumps(profiling_data, indent=2, sort_keys=True))



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
from __future__ import division

import collections
import numbers
import random

import torch

from docling_ibm_models.tableformer.data_management import functional as F


class Normalize(object):
    """Normalize a tensor image with mean and standard deviation.
    Given mean: ``(M1,...,Mn)`` and std: ``(S1,..,Sn)`` for ``n`` channels, this transform
    will normalize each channel of the input ``torch.*Tensor`` i.e.
    ``input[channel] = (input[channel] - mean[channel]) / std[channel]``
    Args:
        mean (sequence): Sequence of means for each channel.
        std (sequence): Sequence of standard deviations for each channel.
    """

    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, tensor, target=None):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
        Returns:
            Tensor: Normalized Tensor image.
        """
        return F.normalize(tensor, self.mean, self.std), target

    def __repr__(self):
        return self.__class__.__name__ + "(mean={0}, std={1})".format(
            self.mean, self.std
        )


class Resize(object):
    """Resize the input PIL Image to the given size.
    Args:
        size (sequence or int): Desired output size. If size is a sequence like
        (h, w), output size will be matched to this. If size is an int,
        smaller edge of the image will be matched to this number.
        i.e, if height > width, then image will be rescaled to
        (size * height / width, size)
        interpolation (int, optional): Desired interpolation. Default is
        ``BILINEAR``
    """

    def __init__(self, size, interpolation="BILINEAR"):
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img, target=None):
        """
        Args:
        img (np.ndarray): Image to be scaled.
        Returns:
        np.ndarray: Rescaled image.
        """
        # Resize bboxes (in pixels)
        x_scale = 0
        y_scale = 0

        if img.shape[1] > 0:
            x_scale = self.size[0] / img.shape[1]
        if img.shape[0] > 0:
            y_scale = self.size[1] / img.shape[0]

        # loop over bboxes
        if target is not None:
            if target["boxes"] is not None:
                target_ = target.copy()
                target_["boxes"][:, 0] = x_scale * target_["boxes"][:, 0]
                target_["boxes"][:, 1] = y_scale * target_["boxes"][:, 1]
                target_["boxes"][:, 2] = x_scale * target_["boxes"][:, 2]
                target_["boxes"][:, 3] = y_scale * target_["boxes"][:, 3]
        return F.resize(img, self.size, self.interpolation), target

    def __repr__(self):
        interpolate_str = self.interpolation
        return self.__class__.__name__ + "(size={0}, interpolation={1})".format(
            self.size, interpolate_str
        )



#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#
import json
import logging
import math
import statistics

import docling_ibm_models.tableformer.settings as s
from docling_ibm_models.tableformer.data_management.tf_cell_matcher import CellMatcher

LOG_LEVEL = logging.INFO
# LOG_LEVEL = logging.DEBUG


class MatchingPostProcessor:
    r"""
    The MatchingPostProcessor aims to improve the matchings between the predicted table cells and
    the pdf cells
    """

    def __init__(self, config):
        self._config = config
        self._cell_matcher = CellMatcher(config)

    def _log(self):
        # Setup a custom logger
        return s.get_custom_logger(self.__class__.__name__, LOG_LEVEL)

    def _get_table_dimension(self, table_cells):
        r"""
        Get dimensions (columns, rows) of a table from table_cells

        Parameters
        ----------
        table_cells : list of dict
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"

        Returns
        -------
        columns : integer,
        rows : integer,
        max_cell_id : integer,
            highest cell_id in table_cells
        """
        columns = 1
        rows = 1
        max_cell_id = 0

        for cell in table_cells:
            if cell["column_id"] > columns:
                columns = cell["column_id"]
            if cell["row_id"] > rows:
                rows = cell["row_id"]
            if cell["cell_id"] > max_cell_id:
                max_cell_id = cell["cell_id"]

        return columns + 1, rows + 1, max_cell_id

    def _get_good_bad_cells_in_column(self, table_cells, column, matches):
        r"""
        1. step
        Get good/bad IOU predicted cells for each structural column (of minimal grid)

        Parameters
        ----------
        table_cells : list of dict
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"
        column : integer
            Index of a column
        matches : dictionary of lists of table_cells
            A dictionary which is indexed by the pdf_cell_id as key and the value is a list
            of the table_cells that fall inside that pdf cell

        Returns
        -------
        good_table_cells : list of dict
            cells in a column that have match
        bad_table_cells : list of dict
            cells in a column that don't have match
        """
        good_table_cells = []
        bad_table_cells = []

        for cell in table_cells:
            if cell["column_id"] == column:
                table_cell_id = cell["cell_id"]

                bad_match = True
                allow_class = True

                for pdf_cell_id in matches:
                    # CHECK IF CELL CLASS TO BE VERIFIED HERE
                    if "cell_class" in cell:
                        if cell["cell_class"] <= 1:
                            allow_class = False
                    else:
                        self._log().debug("***")
                        self._log().debug("no cell_class in...")
                        self._log().debug(cell)
                        self._log().debug("***")
                    if allow_class:
                        match_list = matches[pdf_cell_id]
                        for match in match_list:
                            if match["table_cell_id"] == table_cell_id:
                                good_table_cells.append(cell)
                                bad_match = False
                if bad_match:
                    bad_table_cells.append(cell)

        return good_table_cells, bad_table_cells

    def _delete_column_from_table(self, table_cells, column):
        r"""
        1.a. step
        If all IOU in a column are bad - eliminate column (from bboxes and structure)

        Parameters
        ----------
        table_cells : list of dict
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"
        column : integer
            Index of a column

        Returns
        -------
        new_table_cells : list of dict
        """
        new_table_cells = []

        for cell in table_cells:
            if cell["column_id"] < column:
                new_table_cells.append(cell)
            if cell["column_id"] > column:
                new_cell = {
                    "bbox": cell["bbox"],
                    "cell_id": cell["cell_id"],
                    "column_id": cell["column_id"] - 1,
                    "label": cell["label"],
                    "row_id": cell["row_id"],
                    "cell_class": cell["cell_class"],
                }
                new_table_cells.append(new_cell)

        return new_table_cells

    def _find_alignment_in_column(self, cells):
        r"""
        2. step
        Find alignment of good IOU cells per column

        Parameters
        ----------
        cells : list of dict
            Cells in a column
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"

        Returns
        -------
        alignment : string
            column general alignment can be: "left", "right", "center"
        """
        possible_alignments = ["left", "middle", "right"]
        alignment = "left"  # left / right / center

        lefts = []
        rights = []
        middles = []

        for cell in cells:
            x_left = cell["bbox"][0]
            x_right = cell["bbox"][2]
            x_middle = (x_left + x_right) / 2
            lefts.append(x_left)
            rights.append(x_right)
            middles.append(x_middle)

        if len(lefts) > 0:
            delta_left = max(lefts) - min(lefts)
            delta_middle = max(middles) - min(middles)
            delta_right = max(rights) - min(rights)

            deltas = [delta_left, delta_middle, delta_right]
            align_index = deltas.index(min(deltas))
            alignment = possible_alignments[align_index]

        return alignment

    def _get_median_pos_size(self, cells, alignment):
        r"""
        3. step
        Get median* (according to alignment) "bbox left/middle/right X" coord
        for good IOU cells, get median* cell size in a column.

        Parameters
        ----------
        cells : list of dict
            Cells in a column
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"
        alignment : string
            column general alignment can be: "left", "right", "center"

        Returns
        -------
        median_x : number
            Median X position of a cell (according to alignment)
        median_y : number
            Median Y position of a cell (according to alignment)
        median_width : number
            Median width of a cell
        median_height : number
            Median height of a cell
        """
        median_x = 0
        median_y = 0
        median_width = 1
        median_height = 1

        coords_x = []
        coords_y = []
        widths = []
        heights = []

        for cell in cells:
            if "rowspan_val" not in cell:
                if "colspan_val" not in cell:
                    if cell["cell_class"] > 1:
                        # Use left alignment
                        x_coord = cell["bbox"][0]
                        if alignment == "middle":
                            # Use middle alignment
                            x_coord = (cell["bbox"][2] + cell["bbox"][0]) / 2
                        if alignment == "right":
                            # Use right alignment
                            x_coord = cell["bbox"][2]

                        coords_x.append(x_coord)
                        y_coord = cell["bbox"][1]
                        coords_y.append(y_coord)

                        width = cell["bbox"][2] - cell["bbox"][0]
                        widths.append(width)
                        height = cell["bbox"][3] - cell["bbox"][1]
                        heights.append(height)
                    else:
                        self._log().debug("Empty cells not considered in medians")
                        self._log().debug(cell)
                else:
                    self._log().debug("Colspans not considered in medians")
                    self._log().debug(cell)
            else:
                self._log().debug("Rowspans not considered in medians")
                self._log().debug(cell)

        if len(coords_x) > 0:
            median_x = statistics.median(coords_x)
        if len(coords_y) > 0:
            median_y = statistics.median(coords_y)
        if len(widths) > 0:
            median_width = statistics.median(widths)
        if len(heights) > 0:
            median_height = statistics.median(heights)
        return median_x, median_y, median_width, median_height

    def _move_cells_to_left_pos(
        self, cells, median_x, rescale, median_width, median_height, alignment
    ):
        r"""
        4. step
        Move bad cells to the median* (left/middle/right) good in a column
        (Additionally), re-size cell to median* size of cells in a column

        Parameters
        ----------
        cells : list of dict
            Cells in a column
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"
        median_x : number
            Median X position of a cell (according to alignment)
        rescale : boolean
            should cells be re-sized to median or not
        median_width : number
            Median width of a cell
        median_height : number
            Median height of a cell
        alignment : string
            column general alignment can be: "left", "right", "center"

        Returns
        -------

        new_table_cells : list of dict
            Cells in a column
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"
        """
        new_table_cells = []

        for cell in cells:
            new_cell = {
                "bbox": [],
                "cell_id": 0,
                "column_id": 0,
                "label": "",
                "row_id": 0,
                "cell_class": 0,
            }
            x1 = cell["bbox"][0]
            y1 = cell["bbox"][1]
            x2 = cell["bbox"][2]
            y2 = cell["bbox"][3]
            original_width = x2 - x1
            # original_height = y2 - y1

            # Move to left by default
            new_x1 = median_x
            new_y1 = y1
            new_x2 = median_x + original_width
            new_y2 = y2

            if rescale:
                new_x2 = median_x + median_width
                # Next line does vertical resizing of BBOX:
                new_y2 = y1 + median_height

            # Move to middle
            if alignment == "middle":
                # TODO
                new_x1 = median_x - (original_width / 2)
                new_x2 = new_x1 + original_width
                if rescale:
                    new_x1 = median_x - (median_width / 2)
                    new_x2 = median_x + (median_width / 2)

            # Move to right
            if alignment == "right":
                new_x1 = median_x - original_width
                new_x2 = median_x
                if rescale:
                    new_x1 = median_x - median_width

            new_cell["bbox"] = [new_x1, new_y1, new_x2, new_y2]
            new_cell["cell_id"] = cell["cell_id"]
            new_cell["column_id"] = cell["column_id"]
            new_cell["label"] = cell["label"]
            new_cell["row_id"] = cell["row_id"]
            new_cell["cell_class"] = cell["cell_class"]
            # Add spans if present
            if "rowspan_val" in cell:
                new_cell["rowspan_val"] = cell["rowspan_val"]
            if "colspan_val" in cell:
                new_cell["colspan_val"] = cell["colspan_val"]
            new_table_cells.append(new_cell)
        return new_table_cells

    def _run_intersection_match(self, cell_matcher, table_cells, pdf_cells):
        r"""
        5. step
        Generate new matches, run Intersection over cell(pdf) on a table cells

        Parameters
        ----------
        cell_matcher : CellMatcher
            src.data_management.cell_matcher
        table_cells : list of dict
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"
        pdf_cells : list of dict
            List of PDF cells as defined by Docling

        Returns
        -------
        clean_matches : dictionary of lists of table_cells
            A dictionary which is indexed by the pdf_cell_id as key and the value is a list
            of the table_cells that fall inside that pdf cell
        """
        new_matches = {}
        clean_matches = {}
        new_matches, matches_counter = cell_matcher._intersection_over_pdf_match(
            table_cells, pdf_cells
        )
        clean_matches = new_matches
        # Convert to JSON and back to have string keys in the dictionary
        clean_matches_string = json.dumps(clean_matches)
        clean_matches = json.loads(clean_matches_string)
        return clean_matches

    def _find_overlapping(self, table_cells):

        def correct_overlap(box1, box2):
            # Extract coordinates from the bounding boxes
            x1_min, y1_min, x1_max, y1_max = box1["bbox"]
            x2_min, y2_min, x2_max, y2_max = box2["bbox"]

            # Calculate the overlap in both x and y directions
            overlap_x = min(x1_max, x2_max) - max(x1_min, x2_min)
            overlap_y = min(y1_max, y2_max) - max(y1_min, y2_min)

            # If there is no overlap, return the original boxes
            if overlap_x <= 0 or overlap_y <= 0:
                return box1, box2

            # Decide how to push the boxes apart
            if overlap_x < overlap_y:
                # Push horizontally
                if x1_min < x2_min:
                    # Move box1 to the left and box2 to the right
                    box1["bbox"][2] -= math.ceil(overlap_x / 2) + 2
                    box2["bbox"][0] += math.floor(overlap_x / 2)
                else:
                    # Move box2 to the left and box1 to the right
                    box2["bbox"][2] -= math.ceil(overlap_x / 2) + 2
                    box1["bbox"][0] += math.floor(overlap_x / 2)
            else:
                # Push vertically
                if y1_min < y2_min:
                    # Move box1 up and box2 down
                    box1["bbox"][3] -= math.ceil(overlap_y / 2) + 2
                    box2["bbox"][1] += math.floor(overlap_y / 2)
                else:
                    # Move box2 up and box1 down
                    box2["bbox"][3] -= math.ceil(overlap_y / 2) + 2
                    box1["bbox"][1] += math.floor(overlap_y / 2)

            # Will flip coordinates in proper order, if previous operations reversed it
            box1["bbox"] = [
                min(box1["bbox"][0], box1["bbox"][2]),
                min(box1["bbox"][1], box1["bbox"][3]),
                max(box1["bbox"][0], box1["bbox"][2]),
                max(box1["bbox"][1], box1["bbox"][3]),
            ]
            box2["bbox"] = [
                min(box2["bbox"][0], box2["bbox"][2]),
                min(box2["bbox"][1], box2["bbox"][3]),
                max(box2["bbox"][0], box2["bbox"][2]),
                max(box2["bbox"][1], box2["bbox"][3]),
            ]

            return box1, box2

        def do_boxes_overlap(box1, box2):
            B1 = box1["bbox"]
            B2 = box2["bbox"]
            if (
                (B1[0] >= B2[2])
                or (B1[2] <= B2[0])
                or (B1[3] <= B2[1])
                or (B1[1] >= B2[3])
            ):
                return False
            else:
                return True

        def find_overlapping_pairs_indexes(bboxes):
            overlapping_indexes = []
            # Compare each box with every other box (combinations)
            for i in range(len(bboxes)):
                for j in range(i + 1, len(bboxes)):
                    if i != j:
                        if bboxes[i] != bboxes[j]:
                            if do_boxes_overlap(bboxes[i], bboxes[j]):
                                bboxes[i], bboxes[j] = correct_overlap(
                                    bboxes[i], bboxes[j]
                                )

            return overlapping_indexes, bboxes

        overlapping_indexes, table_cells = find_overlapping_pairs_indexes(table_cells)
        return table_cells

    def _align_table_cells_to_pdf(self, table_cells, pdf_cells, matches):
        """
        Align table cell bboxes with good matches to encapsulate matching pdf cells
        """
        pdf_cell_dict = {pdf_cell["id"]: pdf_cell["bbox"] for pdf_cell in pdf_cells}
        table_cell_dict = {cell["cell_id"]: cell for cell in table_cells}

        # Track unique cells we're going to add
        processed_cells = set()

        # First pass - create initial new_table_cells with aligned bboxes
        new_table_cells = []

        for pdf_cell_id, match_list in matches.items():
            # Extract unique table cell ids from match_list
            table_cell_ids = set(int(match["table_cell_id"]) for match in match_list)

            # Get bbox of pdf_cell
            pdf_cell_bbox = pdf_cell_dict.get(int(pdf_cell_id))
            if not pdf_cell_bbox:
                continue

            # Process each unique table cell
            for cell_id in table_cell_ids:
                if cell_id in processed_cells:
                    continue

                table_cell = table_cell_dict.get(cell_id)
                if not table_cell:
                    continue

                # Create new table cell with aligned bbox
                new_table_cell = table_cell.copy()
                new_table_cell["bbox"] = list(pdf_cell_bbox)

                # Set cell class
                if "cell_class" not in new_table_cell:
                    new_table_cell["cell_class"] = "2"

                new_table_cells.append(new_table_cell)
                processed_cells.add(cell_id)

        # Second pass - aggregate bboxes for duplicate cells
        cell_to_bboxes = {}
        for cell in new_table_cells:
            cell_id = cell["cell_id"]
            if cell_id not in cell_to_bboxes:
                cell_to_bboxes[cell_id] = []
            cell_to_bboxes[cell_id].append(cell["bbox"])

        # Create final clean table cells
        clean_table_cells = []
        processed_ids = set()

        for cell in new_table_cells:
            cell_id = cell["cell_id"]
            if cell_id in processed_ids:
                continue

            bboxes = cell_to_bboxes[cell_id]
            if len(bboxes) > 1:
                # Merge bboxes
                x1s = [bbox[0] for bbox in bboxes]
                y1s = [bbox[1] for bbox in bboxes]
                x2s = [bbox[2] for bbox in bboxes]
                y2s = [bbox[3] for bbox in bboxes]

                cell["bbox"] = [min(x1s), min(y1s), max(x2s), max(y2s)]

            clean_table_cells.append(cell)
            processed_ids.add(cell_id)

        return clean_table_cells

    def _deduplicate_cells(self, tab_columns, table_cells, iou_matches, ioc_matches):
        r"""
        7. step

        De-duplicate columns in table_cells according to highest column score
        in: matches + intersection_pdf_matches

        Parameters
        ----------
        tab_columns : integer
            Number of table columns
        table_cells : list of dict
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"
        iou_matches : dictionary of lists of table_cells
            Cell matches done using Intersection Over Union (IOU) method
        ioc_matches : dictionary of lists of table_cells
            Cell matches done using Intersection Over (PDF) Cell method

        Returns
        -------
        new_table_cells : list of dict
            New table cells with removed column duplicates
        new_matches : dictionary of lists of table_cells
            Matches that are in sync with new_table_cells
        new_tab_columns : integer
            New number of table columns
        """
        pdf_cells_in_columns = []
        total_score_in_columns = []

        for col in range(tab_columns):
            column_table_cells = []
            column_pdf_cells_iou = []
            column_pdf_cells_ioc = []
            column_pdf_cells = []
            column_iou_score = 0
            column_ioc_score = 0

            for cell in table_cells:
                if cell["column_id"] == col:
                    table_cell_id = cell["cell_id"]
                    column_table_cells.append(table_cell_id)

            # SUM IOU + IOC Scores for column, collect all pdf_cell_id
            for iou_key in iou_matches:
                iou_match_list = iou_matches[iou_key]
                for uk in range(len(iou_match_list)):
                    t_cell_id = iou_match_list[uk]["table_cell_id"]
                    if t_cell_id in column_table_cells:
                        if "iou" in iou_match_list[uk]:
                            # In case initial match was IOU
                            column_iou_score += iou_match_list[uk]["iou"]
                        elif "iopdf" in iou_match_list[uk]:
                            # Otherwise it's intersection over PDF match
                            column_iou_score += iou_match_list[uk]["iopdf"]
                        column_pdf_cells_iou.append(iou_key)

            for ioc_key in ioc_matches:
                ioc_match_list = ioc_matches[ioc_key]
                for k in range(len(ioc_match_list)):
                    t_cell_id = ioc_match_list[k]["table_cell_id"]
                    if t_cell_id in column_table_cells:
                        column_ioc_score += ioc_match_list[k]["iopdf"]
                        column_pdf_cells_ioc.append(ioc_key)

            column_pdf_cells = column_pdf_cells_iou
            column_pdf_cells += list(
                set(column_pdf_cells_ioc) - set(column_pdf_cells_iou)
            )
            column_total_score = column_iou_score + column_ioc_score

            pdf_cells_in_columns.append(column_pdf_cells)
            total_score_in_columns.append(column_total_score)
            self._log().debug(
                "Column: {}, Score:{}, PDF cells: {}".format(
                    col, column_total_score, column_pdf_cells
                )
            )

        # Eliminate duplicates in the pdf_cells_in_columns and ensure int content
        # pdf_cells_in_columns:
        # - initially:  list of lists of str with duplicates in the inner lists
        # - afterwards: list of lists of int (unique)
        pdf_cells_in_columns = [
            list(set([x for x in map(lambda x: int(x), le)]))
            for le in pdf_cells_in_columns
        ]
        cols_to_eliminate = []
        # Pairwise comparison of all columns, finding intersection, and it's length
        for cl in range(tab_columns - 1):
            col_a = pdf_cells_in_columns[cl]
            col_b = pdf_cells_in_columns[cl + 1]
            score_a = total_score_in_columns[cl]
            score_b = total_score_in_columns[cl + 1]
            intsct = list(set(col_a).intersection(col_b))
            int_prc = 0
            if len(col_a) > 0:
                int_prc = len(intsct) / len(col_a)
            logstring = "Col A: {}, Col B: {}, Int: {}, %: {}, Score A: {}, Score B: {}"
            self._log().debug(
                logstring.format(cl, cl + 1, len(intsct), int_prc, score_a, score_b)
            )

            # Consider structural column elimination
            # if 60% of two columns pointing to the same pdf cells
            if int_prc > 0.6:
                if score_a >= score_b:
                    # Elliminate B
                    cols_to_eliminate.append(cl + 1)
                if score_b > score_a:
                    # Elliminate A
                    cols_to_eliminate.append(cl)

        self._log().debug("Columns to eliminate: {}".format(cols_to_eliminate))
        new_table_cells = []
        new_matches = {}

        removed_table_cell_ids = []
        new_tab_columns = tab_columns - len(cols_to_eliminate)

        # Clean table_cells structure
        for tab_cell in table_cells:
            add_cell = True
            for col_del in cols_to_eliminate:
                if tab_cell["column_id"] == col_del:
                    removed_table_cell_ids.append(tab_cell["cell_id"])
                    add_cell = False
            if add_cell:
                new_table_cells.append(tab_cell)
        # Clean ioc_matches structure
        for pdf_cell_id, pdf_cell_matches in ioc_matches.items():
            new_cell_match = []
            for pdf_match in pdf_cell_matches:
                if pdf_match["table_cell_id"] not in removed_table_cell_ids:
                    new_cell_match.append(pdf_match)

            if len(new_cell_match) > 0:
                new_matches[pdf_cell_id] = new_cell_match

        return new_table_cells, new_matches, new_tab_columns

    def _do_final_asignment(self, table_cells, iou_matches, ioc_matches):
        r"""
        8. step

        Do final assignment of table bbox to pdf cell based on saved scores,
        either preferring IOU over PDF Intersection, and higher Intersection over lower,
        or just use PDF Intersection
        Rule: 1 Table cell can contain many PDF cells,
            but each PDF cell has to be asigned to one Table cell
        Rule: Do not discard table bboxes at this point, asign all of them

        Iterate over matches, if PDF cell has more than 1 table cell match:
        Go over all other matches and delete tab_cell match of lower score
        (prefer iou match over ioc match)

        Parameters
        ----------
        table_cells : list of dict
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"
        iou_matches : dictionary of lists of table_cells
            Cell matches done using Intersection Over Union (IOU) method
        ioc_matches : dictionary of lists of table_cells
            Cell matches done using Intersection Over (PDF) Cell method

        Returns
        -------
        new_matches : dictionary of lists of table_cells
            New matches with final table cell asignments
        """
        new_matches = {}

        for pdf_cell_id, pdf_cell_matches in ioc_matches.items():
            max_ioc_match = max(pdf_cell_matches, key=lambda x: x["iopdf"])
            new_matches[pdf_cell_id] = [max_ioc_match]

        return new_matches

    def _merge_two_bboxes(self, bbox1, bbox2):
        r"""
        Merge two bboxes into one bboxes that encompasses the two

        Parameters
        ----------
        bbox1 : list of numbers
            bbox to be merged described as two corners [x1, y1, x2, y2]
        bbox1 : list of numbers
            bbox to be merged described as two corners [x1, y1, x2, y2]

        Returns
        -------
        bbox_result : list of numbers
            bbox that encompasses two input bboxes
        """
        bbox_result = [-1, -1, -1, -1]
        bbox_result[0] = min([bbox1[0], bbox2[0]])
        bbox_result[1] = min([bbox1[1], bbox2[1]])
        bbox_result[2] = max([bbox1[2], bbox2[2]])
        bbox_result[3] = max([bbox1[3], bbox2[3]])
        return bbox_result

    def _pick_orphan_cells(
        self, tab_rows, tab_cols, max_cell_id, table_cells, pdf_cells, matches
    ):
        # 9.
        # new_matches, new_table_cells, max_cell_id
        r"""
        9. step

        Pick up remaining orphan cells (pdf cells that don't have any matches or intersections)
            9.a. Determine row banding per row (min/max Y per row)
                match with orphan cells (intersection with band)
            9.b. Determine column banding per column (min/max X per column)
                match with orphan cells (intersection with band)
            9.c. Decide to which column/row each orphan PDF cell belongs
                append match to existing structural cell, or create cell

        Parameters
        ----------
        tab_rows : number
            Total number of rows
        tab_cols : number
            Total number of columns
        max_cell_id : number
            Highest table cell id
        table_cells : list of dict
            Each value is a dictionary with keys: "cell_id", "row_id", "column_id", "bbox", "label"
        pdf_cells : list of dict
            List of PDF cells as defined by Docling
        matches : dictionary of lists of table_cells
            A dictionary which is indexed by the pdf_cell_id as key and the value is a list
            of the table_cells that fall inside that pdf cell

        Returns
        -------
        new_matches : dictionary of lists of table_cells
            updated matches
        new_table_cells : list of dict
            updated table cells
        max_cell_id : number
            New highest table cell id, accounting freshly added table cells (if any)
        """

        new_matches = matches
        new_table_cells = table_cells

        # Identify orphan rows (START)
        orphan_rows = []
        orphan_rows_depth = []
        orphan_rows_bbox = []

        # List with pdf_ids which are used in some (any) row
        used_row_pdf_ids = []
        used_row_rowid = []

        for row in range(tab_rows):
            bbox_y1s = []  # y2 > y1
            bbox_y2s = []
            row_y1 = -1
            row_y2 = -1
            row_table_cells = []
            row_table_cell_ids = []
            orphan_cells_in_row = []
            orphan_cells_in_row_depth = []
            orphan_cells_in_row_bbox = []

            for cell in table_cells:
                if cell["row_id"] == row:
                    # Do not consider spanned cells
                    if "rowspan_val" not in cell:
                        # Do not consider empty cells
                        if cell["cell_class"] > 1:
                            table_cell_id = cell["cell_id"]
                            row_table_cells.append(cell)
                            row_table_cell_ids.append(table_cell_id)
                            bbox_y1s.append(cell["bbox"][1])
                            bbox_y2s.append(cell["bbox"][3])

            # Y coordinates that define band of rows
            if len(bbox_y1s) > 0:
                row_y1 = min(bbox_y1s)
            if len(bbox_y2s) > 0:
                row_y2 = max(bbox_y2s)

            # Find "orphan" cells that intersect the band
            for pdf_cell in pdf_cells:
                pdf_str_id = str(pdf_cell["id"])
                if pdf_str_id not in matches:
                    within_band = False
                    depth = -1

                    centroid_band = (row_y2 + row_y1) / 2
                    centroid_cell = (pdf_cell["bbox"][3] + pdf_cell["bbox"][1]) / 2

                    # pdf_cell - Orphan, and don't have any match
                    if pdf_cell["bbox"][1] >= row_y1 and pdf_cell["bbox"][1] <= row_y2:
                        depth = abs(centroid_band - centroid_cell)
                        within_band = True

                    if pdf_cell["bbox"][3] >= row_y1 and pdf_cell["bbox"][3] <= row_y2:
                        depth = abs(centroid_band - centroid_cell)
                        within_band = True

                    if pdf_cell["bbox"][1] <= row_y1 and pdf_cell["bbox"][3] >= row_y2:
                        depth = abs(centroid_band - centroid_cell)
                        within_band = True

                    if within_band:
                        if pdf_str_id not in used_row_pdf_ids:
                            used_row_pdf_ids.append(pdf_str_id)
                            used_row_rowid.append(row)
                            orphan_cells_in_row.append(pdf_str_id)
                            orphan_cells_in_row_depth.append(round(depth))
                            orphan_cells_in_row_bbox.append(pdf_cell["bbox"])
                        else:
                            self._log().debug("Found duplicate: {}".format(pdf_str_id))
                            # Get index of a row where pdf cell was already detected
                            used_ind = used_row_pdf_ids.index(pdf_str_id)
                            duplicate_id_found_in_row = used_row_rowid[used_ind]
                            valid_duplicate = False
                            if len(orphan_rows) > duplicate_id_found_in_row:
                                index_in_row_list = orphan_rows[
                                    duplicate_id_found_in_row
                                ].index(pdf_str_id)
                                valid_duplicate = True

                            if valid_duplicate:
                                i1 = duplicate_id_found_in_row
                                i2 = index_in_row_list
                                score1 = orphan_rows_depth[i1][i2]
                                score2 = round(depth)
                                # If new cell better than the old one
                                if score2 < score1:
                                    # Delete old record about the pdf cell...
                                    orphan_rows[i1].pop(index_in_row_list)
                                    orphan_rows_depth[i1].pop(index_in_row_list)
                                    orphan_rows_bbox[i1].pop(index_in_row_list)

                                    used_row_pdf_ids.pop(used_ind)
                                    used_row_rowid.pop(used_ind)
                                    # Then proceed adding new cell
                                    used_row_pdf_ids.append(pdf_str_id)
                                    used_row_rowid.append(row)
                                    orphan_cells_in_row.append(pdf_str_id)
                                    orphan_cells_in_row_depth.append(round(depth))
                                    orphan_cells_in_row_bbox.append(pdf_cell["bbox"])
                                    msg = "Resolved duplicate: {} in favor of new one"
                                    self._log().debug(msg.format(pdf_str_id))
                                else:
                                    msg = "Resolved duplicate: {} in favor of old one"
                                    self._log().debug(msg.format(pdf_str_id))

            log_msg = "Row: {}, Band: {}/{}, Orphan PDF cells: {}"
            self._log().debug(log_msg.format(row, row_y1, row_y2, orphan_cells_in_row))
            orphan_rows.append(orphan_cells_in_row)
            orphan_rows_depth.append(orphan_cells_in_row_depth)
            orphan_rows_bbox.append(orphan_cells_in_row_bbox)

        # Identify orphan rows (END)
        self._log().debug("...")
        # Identify orphan columns
        orphan_columns = []
        orphan_columns_depth = []
        orphan_columns_bbox = []
        used_col_pdf_ids = []
        used_col_columnid = []

        for col in range(tab_cols):
            bbox_x1s = []  # y2 > y1
            bbox_x2s = []
            col_x1 = -1
            col_x2 = -1
            col_table_cells = []
            col_table_cell_ids = []
            orphan_cells_in_col = []
            orphan_cells_in_col_depth = []
            orphan_cells_in_col_bbox = []

            for cell in table_cells:
                if cell["column_id"] == col:
                    # Do not consider spanned cells
                    if "colspan_val" not in cell:
                        # Do not consider empty cells
                        if cell["cell_class"] > 1:
                            table_cell_id = cell["cell_id"]
                            col_table_cells.append(cell)
                            col_table_cell_ids.append(table_cell_id)
                            bbox_x1s.append(cell["bbox"][0])
                            bbox_x2s.append(cell["bbox"][2])
                    else:
                        wrn_txt = "Orphan matching skipped cell in column {} because of colspan"
                        self._log().debug(wrn_txt.format(col))
                        # self._log().info(cell)

            # X coordinates that define band of columns
            if len(bbox_x1s) > 0:
                col_x1 = min(bbox_x1s)
            if len(bbox_x2s) > 0:
                col_x2 = max(bbox_x2s)

            # Find "orphan" cells that intersect the band
            for pdf_cell in pdf_cells:
                pdf_str_id = str(pdf_cell["id"])
                if pdf_str_id not in matches:
                    within_band = False
                    depth = -1

                    centroid_band = (col_x2 + col_x1) / 2
                    centroid_cell = (pdf_cell["bbox"][2] + pdf_cell["bbox"][0]) / 2

                    if pdf_cell["bbox"][0] >= col_x1 and pdf_cell["bbox"][0] <= col_x2:
                        depth = abs(centroid_band - centroid_cell)
                        within_band = True

                    if pdf_cell["bbox"][2] >= col_x1 and pdf_cell["bbox"][2] <= col_x2:
                        depth = abs(centroid_band - centroid_cell)
                        within_band = True

                    if pdf_cell["bbox"][0] < col_x1 and pdf_cell["bbox"][2] > col_x2:
                        depth = abs(centroid_band - centroid_cell)
                        within_band = True

                    if within_band:
                        if pdf_str_id not in used_col_pdf_ids:
                            used_col_pdf_ids.append(pdf_str_id)
                            used_col_columnid.append(col)
                            orphan_cells_in_col.append(pdf_str_id)
                            orphan_cells_in_col_depth.append(round(depth))
                            orphan_cells_in_col_bbox.append(pdf_cell["bbox"])
                        else:
                            self._log().debug("Found duplicate: {}".format(pdf_str_id))
                            # Get index of a column where pdf cell was already detected
                            used_ind = used_col_pdf_ids.index(pdf_str_id)
                            duplicate_id_found_in_column = used_col_columnid[used_ind]

                            valid_col_duplicate = False
                            if len(orphan_columns) > duplicate_id_found_in_column:
                                index_in_col_list = orphan_columns[
                                    duplicate_id_found_in_column
                                ].index(pdf_str_id)
                                valid_col_duplicate = True

                            if valid_col_duplicate:
                                i1 = duplicate_id_found_in_column
                                i2 = index_in_col_list
                                score1 = orphan_columns_depth[i1][i2]
                                score2 = round(depth)
                                # If new cell better than the old one
                                if score2 < score1:
                                    # Delete old record about the pdf cell...
                                    orphan_columns[i1].pop(index_in_col_list)
                                    orphan_columns_depth[i1].pop(index_in_col_list)
                                    orphan_columns_bbox[i1].pop(index_in_col_list)

                                    used_col_pdf_ids.pop(used_ind)
                                    used_col_columnid.pop(used_ind)
                                    # Then proceed adding new cell
                                    used_col_pdf_ids.append(pdf_str_id)
                                    used_col_columnid.append(col)
                                    orphan_cells_in_col.append(pdf_str_id)
                                    orphan_cells_in_col_depth.append(round(depth))
                                    orphan_cells_in_col_bbox.append(pdf_cell["bbox"])
                                    msg = "Resolved duplicate: {} in favor of new one"
                                    self._log().debug(msg.format(pdf_str_id))
                                else:
                                    msg = "Resolved duplicate: {} in favor of old one"
                                    self._log().debug(msg.format(pdf_str_id))

            orphan_columns.append(orphan_cells_in_col)
            orphan_columns_depth.append(orphan_cells_in_col_depth)
            orphan_columns_bbox.append(orphan_cells_in_col_bbox)

        # Assign to structural cells and/or create new cells when absent

        for col_ind in range(len(orphan_columns)):
            self._log().debug(
                "Col: {}, Orphan PDF cells: {}".format(col_ind, orphan_columns[col_ind])
            )
            self._log().debug(
                "Col: {},     Orphan Depth: {}".format(
                    col_ind, orphan_columns_depth[col_ind]
                )
            )
        self._log().debug("...")

        # Collect the pdf_ids from the orphan_rows and sort them in order to produce the same
        # results with the c++ implementation
        orphan_rows_pdf_ids = []
        row_id_per_pdf_id = {}  # pdf_cell_id -> row_id
        for row_id, row_pdf_ids in enumerate(orphan_rows):
            if len(row_pdf_ids) == 0:
                continue
            # Extend the orphan_rows_pdf_ids with the pdf_ids as int
            orphan_rows_pdf_ids.extend([int(x) for x in row_pdf_ids])
            # Set the row_id for the pdf_ids
            for pdf_cell_id in row_pdf_ids:
                row_id_per_pdf_id[int(pdf_cell_id)] = row_id

        orphan_rows_pdf_ids.sort()

        # Assign Table cell Row ID / Table cell Column ID to orphans,
        # Check if Table cell doesn't exist in the table_cells, create one,
        # add match to new_matches
        for pdf_cell_id_int in orphan_rows_pdf_ids:
            new_row_id = row_id_per_pdf_id[pdf_cell_id_int]
            new_column_id = 0
            pdf_cell_id = str(pdf_cell_id_int)

            if pdf_cell_id in used_col_pdf_ids:
                new_column_id = used_col_columnid[used_col_pdf_ids.index(pdf_cell_id)]

                self._log().debug(
                    "new_column_id {}, pdf_cell_id {}".format(
                        new_column_id, pdf_cell_id
                    )
                )
                self._log().debug(orphan_columns[new_column_id])
                depth_index = orphan_columns[new_column_id].index(pdf_cell_id)
                confidence = orphan_columns_depth[new_column_id][depth_index]
                pdf_bbox = orphan_columns_bbox[new_column_id][depth_index]

                # 1. Find table_cell_id by new_row_id / new_column_id
                new_table_cell_id = -1
                tcell = list(
                    filter(
                        lambda table_cell: table_cell["row_id"] == new_row_id
                        and table_cell["column_id"] == new_column_id,
                        table_cells,
                    )
                )

                if len(tcell) > 0:
                    new_table_cell_id = tcell[0]["cell_id"]
                    self._log().debug(
                        "reusing table_cell_id: {}".format(new_table_cell_id)
                    )

                    for i in range(len(new_table_cells)):
                        if new_table_cells[i]["cell_id"] == new_table_cell_id:
                            bbox_tmp = self._merge_two_bboxes(
                                new_table_cells[i]["bbox"], pdf_bbox
                            )
                            new_table_cells[i]["bbox"] = bbox_tmp

                if new_table_cell_id < 0:
                    max_cell_id += 1
                    new_table_cell_id = max_cell_id

                    new_table_cell = {
                        "bbox": pdf_bbox,
                        "cell_id": new_table_cell_id,
                        "column_id": new_column_id,
                        "label": "body",
                        "row_id": new_row_id,
                        "cell_class": 2,
                    }
                    self._log().debug(
                        "making new table_cell_id: {}".format(new_table_cell_id)
                    )
                    new_table_cells.append(new_table_cell)

                # And then add new match to the new_matches
                new_matches[str(pdf_cell_id)] = [
                    {"post": confidence, "table_cell_id": new_table_cell_id}
                ]
        return new_matches, new_table_cells, max_cell_id

    def _clear_pdf_cells(self, pdf_cells):
        r"""
        Clean PDF cells from cells that have an empty string as text

        Parameters
        ----------
        pdf_cells : list of dict
            List of PDF cells as defined by Docling

        Returns
        -------
        new_pdf_cells : list of dict
            updated, cleaned list of pdf_cells
        """
        new_pdf_cells = []
        for i in range(len(pdf_cells)):
            if pdf_cells[i]["text"] != "":
                new_pdf_cells.append(pdf_cells[i])
        return new_pdf_cells

    def process(self, matching_details, correct_overlapping_cells=False):
        r"""
        Do post processing, see details in the comments below

        Parameters
        ----------
        matching_details : dictionary
            contains all the necessary information for Docling processing
            already has predictions and initial (IOU) matches

        Returns
        -------
        matching_details : dictionary
            matching_details that contain post-processed matches
        """

        # ====================================================================================
        # Start post-processing
        # ====================================================================================

        # General description of post-processing algorithm to improve matching

        # Uses: (IOU) matches, pdf_cells, table_cells
        # Generates: new_table_cells, new_matches

        # +0. Get minimal grid table dimension (cols/rows)
        # +1. Get good/bad IOU predicted cells for each structural column (of minimal grid)
        #   +1.a. If all IOU in a column are bad - eliminate column (from bboxes and structure)
        # +2. Find alignment of good IOU cells per column
        #   +2.a. For this, measure min-max distance for left, min-max distance for centroid,
        #         min-max distance for right side of cell rectangles
        #         smallest distance would determine Left / Center / Right alignment
        # +3. Get median* (according to alignment) "bbox left/middle/right X" coord
        #         for good IOU cells, get median* cell size in a column.
        # +4. Move bad cells to the median* (left/middle/right) good in a column
        #   +4.a. (Additionally), re-size cell to median* size of cells in a column
        # +5. Generate new matches, run Intersection over cell(pdf) on a table cells

        # NOT USED +6. Align table cell bboxes with good matches
        #              to encapsulate matching pdf cells

        # +7. De-duplicate columns in table_cells according to highest column score
        #        in: matches + intersection_pdf_matches
        # +8. Do final assignment of table bbox to pdf cell based on saved scores,
        #     either preferring IOU over PDF Intersection, and higher Intersection over lower,
        #     or just use PDF Intersection
        #     Rule: 1 Table cell can contain many PDF cells,
        #       but each PDF cell has to be asigned to one Table cell
        #     Rule: Do not discard table bboxes at this point, asign all of them
        #   +8.a. Align table cell bboxes with matched pdf cells bboxes (instead of step 6)

        # +9. Pick up remaining orphan cells
        #   +9.a. Determine row banding per row (min/max Y per row)
        #           match with orphan cells (intersection with band)
        #   +9.b. Determine column banding per column (min/max X per column)
        #           match with orphan cells (intersection with band)
        #   +9.c. Decide to which column/row each orphan PDF cell belongs
        #           append match to existing structural cell,
        #           or create cell
        #  ?9.d. For each pdf-cell from the list find closest and MATCH with row-table-cell
        #        (minimal rectangle to-rectangle distance)

        # *Use median instead of average to account for occasional
        # colspans that would span across columns,
        # as they are rare they shouldn't indfluence much a median position
        # of other cells in a minimal-grid column

        self._log().debug("Start prediction post-processing...")
        table_cells = matching_details["table_cells"]
        pdf_cells = self._clear_pdf_cells(matching_details["pdf_cells"])
        matches = matching_details["matches"]

        # ------------------------------------------------------------------------------------------
        # -1. If initial (IOU) matches are empty,
        # generate new ones based on intersection over cell

        if not matches:
            self._log().debug(
                "-----------------------------------------------------------------"
            )
            self._log().debug(
                "-----------------------------------------------------------------"
            )
            self._log().debug(
                "-   NO INITIAL MATCHES TO POST PROCESS, GENERATING NEW ONES...  -"
            )
            self._log().debug(
                "-----------------------------------------------------------------"
            )
            self._log().debug(
                "-----------------------------------------------------------------"
            )
            matches = self._run_intersection_match(
                self._cell_matcher, table_cells, pdf_cells
            )

        # ------------------------------------------------------------------------------------------
        # 0. Get minimal grid table dimension (cols/rows)
        tab_columns, tab_rows, max_cell_id = self._get_table_dimension(table_cells)
        self._log().debug(
            "COLS {}/ ROWS {}/ MAX CELL ID {}".format(
                tab_columns, tab_rows, max_cell_id
            )
        )

        good_table_cells = []
        bad_table_cells = []
        new_bad_table_cells = []
        fixed_table_cells = []

        # 1. Get good/bad IOU predicted cells for each structural column (of minimal grid)
        for col in range(tab_columns):
            g1, g2 = self._get_good_bad_cells_in_column(table_cells, col, matches)
            good_table_cells = g1
            bad_table_cells = g2
            self._log().debug(
                "COLUMN {}, Good table cells: {}".format(col, len(good_table_cells))
            )
            self._log().debug(
                "COLUMN {}, Bad table cells: {}".format(col, len(bad_table_cells))
            )

            # 2. Find alignment of good IOU cells per column
            alignment = self._find_alignment_in_column(good_table_cells)
            self._log().debug("COLUMN {}, Alignment: {}".format(col, alignment))
            # alignment = "left"

            # 3. Get median (according to alignment) "bbox left/middle/right X"
            #    coordinate for good IOU cells, get median* cell size in a column.
            gm1, gm2, gm3, gm4 = self._get_median_pos_size(good_table_cells, alignment)
            median_x = gm1
            # median_y = gm2
            median_width = gm3
            median_height = gm4
            self._log().debug("Median good X = {}".format(median_x))

            # 4. Move bad cells to the median* (left/middle/right) good in a column
            # nc = self._move_cells_to_left_pos(bad_table_cells, median_x, True,
            # TODO:
            nc = self._move_cells_to_left_pos(
                bad_table_cells, median_x, False, median_width, median_height, alignment
            )
            new_bad_table_cells = nc
            fixed_table_cells.extend(good_table_cells)
            fixed_table_cells.extend(new_bad_table_cells)

        # ====================================================================================
        # Sort table_cells by cell_id before running IOU, to have correct indexes on the output
        fixed_table_cells_sorted = sorted(fixed_table_cells, key=lambda k: k["cell_id"])

        # 5. Generate new matches, run Intersection over cell(pdf) on a table cells
        ip = self._run_intersection_match(
            self._cell_matcher, fixed_table_cells_sorted, pdf_cells
        )
        intersection_pdf_matches = ip

        # 6. NOT USED

        # 7. De-duplicate columns in aligned_table_cells
        # according to highest column score in: matches + intersection_pdf_matches
        # (this is easier now, because duplicated cells will have same bboxes)
        dd1, dd2, dd3 = self._deduplicate_cells(
            tab_columns, fixed_table_cells_sorted, matches, intersection_pdf_matches
        )
        dedupl_table_cells = dd1
        dedupl_matches = dd2

        self._log().debug("...")

        # 8. Do final assignment of table bbox to pdf cell based on saved scores,
        # preferring IOU over PDF Intersection, and higher Intersection over lower
        # ! IOU matches currently disabled,
        # and final assigment is done only on IOC matches
        final_matches = self._do_final_asignment(
            dedupl_table_cells, matches, dedupl_matches
        )

        # 8.a. Re-align bboxes / re-run matching
        dedupl_table_cells_sorted = sorted(
            dedupl_table_cells, key=lambda k: k["cell_id"]
        )

        if (
            len(pdf_cells) > 300
        ):  # For performance, skip this step if there are too many pdf_cells
            aligned_table_cells2 = dedupl_table_cells_sorted
        else:
            aligned_table_cells2 = self._align_table_cells_to_pdf(
                dedupl_table_cells_sorted, pdf_cells, final_matches
            )

        # 9. Distance-match orphans
        po1, po2, po3 = self._pick_orphan_cells(
            tab_rows,
            tab_columns,
            max_cell_id,
            aligned_table_cells2,
            pdf_cells,
            final_matches,
        )
        final_matches_wo = po1
        table_cells_wo = po2
        max_cell_id = po3

        if correct_overlapping_cells:
            # As the last step - correct cell bboxes in a way that they don't overlap:
            if len(table_cells_wo) <= 300:  # For performance reasons
                table_cells_wo = self._find_overlapping(table_cells_wo)

        self._log().debug("*** final_matches_wo")
        self._log().debug(final_matches_wo)
        self._log().debug("*** table_cells_wo")
        self._log().debug(table_cells_wo)

        for pdf_cell_id in range(len(final_matches_wo)):
            if str(pdf_cell_id) in final_matches_wo:
                pdf_cell_match = final_matches_wo[str(pdf_cell_id)]
                if len(pdf_cell_match) > 1:
                    l1 = "!!! Multiple - {}x pdf cell match with id: {}"
                    self._log().info(l1.format(len(pdf_cell_match), pdf_cell_id))
                if pdf_cell_match:
                    tcellid = pdf_cell_match[0]["table_cell_id"]
                    for tcell in table_cells_wo:
                        if tcell["cell_id"] == tcellid:
                            mrow = tcell["row_id"]
                            mcol = tcell["column_id"]
                            l2 = "pdf cell: {} -> row: {} | col:{}"
                            self._log().debug(l2.format(pdf_cell_id, mrow, mcol))
            else:
                self._log().debug(
                    "!!! pdf cell doesn't have match: {}".format(pdf_cell_id)
                )

        # Example of an object:
        # matching_details = {
        #    "iou_threshold": self._iou_thres,
        #    "table_bbox": table_bbox,
        #    "prediction": prediction,
        #    "pdf_cells": pdf_cells,
        #    "page_height": docling_table["page_height"],
        #    "page_width": docling_table["page_width"],
        #    "table_cells": table_cells,
        #    "matches": matches
        # }

        matching_details["table_cells"] = table_cells_wo
        matching_details["matches"] = final_matches_wo
        matching_details["pdf_cells"] = pdf_cells

        self._log().debug("Done prediction matching and post-processing!")
        return matching_details
